###############################################################################
# IMPORTS

from classlib.base import *
#import string
###############################################################################
class Man(Obj,Task_handling,Costs_handling):
    """
    Class, containing common methods of management simulation objects.
    The main task is to manipulate the following database tables:

    res[res_id]     Resources (users,carriers, sections...)
    scheds[sched_id] Time schedules for resources...will be implemented later
    tasks[task_id]   Tasks are orders from other management units.
    proc[proc_id]   Processes that are launched to fulfill a task.
    
    """
    def init_obj(self):
        """
        Initialization for management functions.
        It calls the init_obj_specific function. 
        The databeses should be completed by the specific simulation object
        initialization.
        """

        self.output_interval=2.0 # second update

        # time interval in seconds at which the management is  verified
        #self.verify_interval=20.0 # currently not used 

        # time of last verification
        #self.time_verified=self.sim.time_start  # currently not used 
        
        # make simulator to call output method during main cycle
        self.sim.append_obj_output(self)

        # wait with reading and evaluating messages
        # untill output method is called
        self.read_mes_on_receive=0

        # init tasks stuff
        self.init_tasks()
        self.set_state_tasks(self)

        
        # initialize databases
        self.scheds={}
        self.resources={}
                
        # initialize ids
        self.sched_id_count=0 
        #self.res_id_count=0 # defined by hardware
        
        # add operators that can be executed by incoming messages
        # self.mes_ops['']
        self.mes_ops['register']=self.register
        self.mes_ops['unregister']=self.unregister
        
        # the table task_classes, mapping task names to task classes,
        # must be completed within the specific management initialization.
        # Here just the generic task is added
        self.task_classes['task']=Task
        
        # task to query all available destinations
        self.task_classes['poll_destinations']=Poll_destinations
       
        
        # initialize costs data
        self.costs={}

        # initialize specific states
        self.init_obj_specific()
        
        # initialize default costs for management
        self.init_costs()
        
        

    
    def get_state_man(self,state_image):
        """
        Returns state image with management specific states
        """
        state_image=self.get_state_tasks(state_image)

        state_image['output_interval']=self.output_interval
        state_image['resources']=self.resources
        state_image['scheds']=self.scheds
        state_image['sched_id_count']=self.sched_id_count
        state_image['costs']=self.costs
        #state_image['']=self.
       
        return state_image


    def set_state_man(self,state_image):
        """
        Set state image with management specific states
        """
        self.set_state_tasks(self,state_image)
        self.output_interval=state_image['output_interval']
        self.resources=state_image.get('resources',{})
        self.scheds=state_image.get('scheds',{})
        self.sched_id_count=state_image.get('sched_id_count',0)
        self.costs=state_image['costs']
        # self.=state_image['']

    def get_resource_name(self):
        """
        Returns string with name of resource that is managed by this management
        To be overwritten.
        """
        return 'Resource'
        
    def init_costs(self):
        """
        Define default costs here.
        
        To be overwritten.
        """
        self.set_costs(
            {'estim_init':{'name':'Estim. initial costs',
                               'value':80.0*10**3,
                               'mode':'write'},
             'estim_pa':{'name':'Estim. base costs per year',
                         'value':60*10**3,
                         'mode':'write'},
             #'resource_estim_init':{'name':'Estim. initial costs for '+\
             #                       self.get_resource_name()+'s',
             #                       'value':200*10**3,
             #                  'mode':'read'},
             #'resource_estim_pa':{'name':
             #                     'Estim. costs per year for '+\
             #                     self.get_resource_name()+'s',
             #                     'value':200*10**3,
             #                     'mode':'read'},
             #'total_estim_init':{'name':'Estimated initial total costs',
             #                    'value':0.0,
             #                    'mode':'read'},#determined in do_costs_current
             #'total_estim_pa':{'name':'Estimated total costs per year',
             #                  'value':0.0,
             #                  'mode':'read'},# determined in do_costs_current
             },
            lifetime_estim=10)
        self.init_costs_specific()

    def init_costs_specific(self):
        """
        Management specific cost initialization.
        To be overwritten.
        """
        pass
        
    def register(self,message):
        """
        Register a new resource into the management's database.
        """
        res_id=message['from']
        
        # here will be some checks whether res_id is a valid resource

        # initialize dataset in resourcedb 
        self.resources[res_id]={'id':res_id}
        #print 'register:',self.resources[res_id]
        
        # store also rest of the key-value pairs contained in message
        for key in message.keys():
            if (key!='from')&(key!='op'):
                self.resources[res_id][key]=message[key]

        self.register_specific(self.resources[res_id],message)

    def register_specific(self,resource,message={}):
        """
        Management specific resource registration.
        Will be initialization of statistical and cost data.
        """
        pass

    def unregister(self,message):
        """
        Unregister a resource from the management's database.
        """
        res_id=message['from']

        # do some specific preprocessing
        self.unregister_specific(self.resources[res_id],message)
                
        # here should be verified whether this resource is used by processes

        
        
        if self.resources.has_key(res_id):
            self.unregister_specific(self.resources[res_id],message)
            del self.resources[res_id]

        
    def unregister_specific(self,res,message={}):
        """
        Unregister a resource from the management's database.
        """
        pass
            
    def make_panel(self,panel,**kw):
        """
        Make widgets.
        May be overwritten by specific class.
        """
        # initialize widgets with Non objects such that they are not updated
        # during the make phase.
        
        panel.make_standardpanel()

        # add a resource page
        panel.make_page(self.get_resource_name()+'s',
                        self.make_resourcepage,self.update_resourcepage)
        

        # add management specific panel infos
        self.make_panel_specific(panel)


    def make_panel_specific(self,panel):
        """
        Add track element specific panel infos.
        To be overwritten.
        """
        pass

    
    def make_performancepage(self,parent,widgets):
        """
        Default performance gui widget creation.
        Creates widgets in parent frame. All widgets are put in the
        widgets directory with a string as keyword. These widgets
        can be configured in the corresponding update method.

        May be overwritten.
        """
        
        descriptor,dataset=self.format_statistics()
        widgets['statisticsdata']=gui.DatasetEditor(descriptor,dataset,parent)
        widgets['statisticsdata'].pack(expand=1,fill='both')

    def update_performancepage(self,panel,widgets):
        """
        Update of widgets on performance page.
        """
        descriptor,dataset=self.format_statistics()
        widgets['statisticsdata'].update_contents(dataset)
        
    def reset_statistics(self):
        """
        Reset or initialize statistical variables.
        """
        self.reset_tasks()
        self.reset_costs()
        # indicate hat the results of these  task should be ignored
        # because they started in the past
        for task in self.tasks.values():
            task.ignore_accounting()

        self.reset_statistics_specific()
        
    def reset_statistics_specific(self):
        """
        Reset or initialize statistical variables specific to
        a particular type.

        To be overwritten.
        """
        pass

    def format_statistics(self):
        """
        Return descriptor and dataset contents for statistical data
        about the target object.
        """
        #print 'format_statistics for',self.id
            
        descriptor=[{'name':'Number of '+self.get_resource_name()+'s',
                     'key':'resources_n',
                     'type':'integer',
                     'printformat':'%9d',
                     'width':1,
                     'valid':{'max':10**8},
                     'mode':'read'},
                    {'name':'Current number of tasks',
                     'key':'tasks_n',
                     'type':'integer',
                     'printformat':'%9d',
                     'valid':{'min':0,'max':10**8},
                     'mode':'read',
                     'width':0.2,
                     },
                    {'name':'Completed number of tasks',
                     'key':'tasks_completed_n',
                     'type':'integer',
                     'printformat':'%9d',
                     'valid':{'min':0,'max':10**8},
                     'mode':'read',
                     'width':0.2,
                     },
                    ]

        
        dataset={'resources_n':len(self.resources),
                 'tasks_n':len(self.tasks),
                 'tasks_completed_n':self.statistics['tasks_done']}

        self.format_statistics_specific(descriptor,dataset)

        return descriptor,dataset

    def format_statistics_specific(self,descriptor,dataset):
        """
        Format management specific contents for statistical data.
        """
        pass
    

    def reset_costs(self):
        """
        Sets all costs in the costs directory that occurred during
        the current simulation to zero.

        
        """
        if self.costs.has_key('revenue'):
            self.set_cost_value('revenue',0.0)
            self.set_cost_value('revenue_per_h',0.0)
                        
        self.reset_costs_specific()
        
    def reset_costs_specific(self):
        """
        Sets all management specific
        costs in the costs directory that occurred during
        the current simulation to zero.

        
        """
        pass
                

    
        


    def do_costs_current(self):
        """
        Calculates current costs and modifies the costs directory
        
        """
        #self.set_cost_value('total_estim_init',
        #                    self.get_cost_value('base_estim_init'))
        #self.set_cost_value('total_estim_pa',
        #                    self.get_cost_value('base_estim_pa'))

        # do we need to make cost calcs of resources?
        #if self.costs.has_key('per_resource_estim'):
        #    self.do_costs_resources()

        self.do_costs_current_specific()

    #def do_costs_resources(self):
    #    """
    #    Add resource specific costs to total costs in costs dicitonary.
    #    """
    #    # here the quantity dependent costs per resource is simply
    #    # multiplied by the number of resources.
    #    self.add_cost_value('total_estim_init',len(self.resources)*
    #                        self.get_cost_value('per_resource_estim',
    #                                            len(self.resources)))
    #    
    #    self.add_cost_value('total_estim_pa',len(self.resources)*
    #                        self.get_cost_value('per_resource_estim_pa',
    #                                            len(self.resources)))
        
    def do_costs_current_specific(self):
        """
        Calculates management specific current costs and modifies
        the costs dicitionary.
        
        """
        pass
    
    def format_costs_specific(self,quantity,descriptors,table):
        """
        Remove quantity and livetime information from descriptor.
        """
        pass
    
    def make_resourcepage(self,parent,widgets):
        """
        Create common widgets for resources.
        This is mainly a list with resources.
        """
        
        # create resource table
        sel_command=lambda sel_ind=-1,self=self,widgets=widgets:\
                     self._on_show_resource(sel_ind,widgets)
        widgets['resourcetable']=gui.PanedScrolledTable(parent,
                                                    selectcommand=\
                                                                  sel_command,
                                                    #highlightcommand=
                                                    #              sel_command,
                                                    #hullwidth=490,
                                                    #hullheight=200
                                                        )

        
        # fill tables etc. with data
        self.update_resourcepage(parent,widgets)

    
        # pack everything in the right order
        widgets['resourcetable'].pack(expand=1,fill='both')
        
       
   
    def _on_show_resource(self,sel_ind,widgets):
        """
        Show som details of resource in resource details frame.
        """
        pass

        # open widget with more info...
        # res_id=widgets['resourcetable'].get_selection(0)
        # get module pointer to loaded item
        # res=self.obj.sim.split_id(res_id)['mod']
        # widgets['controlpanel'].open_controlpanel(res)
            
    def update_resourcepage(self,parent,widgets):
        """
        Updates all widgets on resource page.
        """
        
        header,table= self.format_resources()
        
        # format more management specific keys for display or add to table
        self.update_resourcepage_specific(parent,widgets,header,table)

        # update table
        widgets['resourcetable'].make_table(header,table)

    def update_resourcepage_specific(self,parent,widgets,header,table):
        """
        Updates all management specific widgets on resource page and table.
        """
        pass

    def format_resources(self):
        """
        Format table with contents for resource data.
        To be overwritten.
        """
        # self.print_state() 
        header=[{'name':self.get_resource_name()+' ID',
                 'key':'id',
                 'type':'text',
                 'width':1,
                 'mode':'read'},]

        # update resource table
        table=[]

        # copy simply values of all keys in the resource dicitionary 
        if len(self.resources.values())>0:
            for res in self.resources.values():
                row={}
                row.update(res)
                table.append(row)

        self.format_resources_specific(header,table)
        return header,table

    def format_resources_specific(self,header,table):
        """
        Format table with management specific contents for resource data.
        To be overwritten.
        """
        pass
    
    def show_info(self,debuglevel=0):
        #print 'self.mes_ops:'
        #for key in self.mes_ops.keys(): print '  ',key

        
        if debuglevel<=self.debug:
            self.show_info_specific(debuglevel)

    def show_info_specific(self,debuglevel):
        pass

    def format_info_specific(self,il,sep=' '):
        """
        Returns a string list with formatted info.
        """
        dataset_func=gui.DataSet()
        descriptors,table=self.format_resources()
        if descriptors:
            il.append('\n Resources:\n')
            il+=dataset_func.format_table(descriptors,table)
###############################################################################
# THE CLASSES Task AND Proc ARE  IN THE BASE LIBRARY
###############################################################################
        

###############################################################################
# Some commonly used tasks and processes
###############################################################################

class Poll_destinations(Task):
    """
    Task to poll the destinations of all track clusters.

    This polling could be done more simple and more direct,
    but it is used to test the task->process mechanism.
    """
    def init_task(self,args=None):
        """
        Search for track managements in the net and start
        a poll process for each.
        """
        # the global information is currently obtained from the
        # simulation environment (which will not exist in reality),
        # instead, a network management needs to be implemented...
        
        # go through all available simulation object types (=keys)
        for obj_type in self.obj.sim.objs.keys():
            if string.find(obj_type,'trackman')!= -1: # is it a trackman ?
                # start poll destinations process for this trackman
                self.add_proc(Poll_dest,obj_type)
                
        # Now wait untill processes return destination
        
    def put_info(self,info,sender_id=''):
        """
        Another task, which has been ordered by this task
        returns information in form of a dictionary.
        The info is read and the status of this task is changed accordingly.

        May be overwritten by specific management object.
        """

        #print 'update_procs: dest_ids=',self.obj.id,info['dest_ids']
        # store destination id (key) together with track management id
        for dest_id in info['dest_ids']:
            self.obj.dests[dest_id]=info['trackman_id']
            
        # Here the status of the task is simply set to the status of
        # the task which had been ordered.
        self.set_status(info['status'])

        
    def update_procs(self,proc):
        """
        Extend management's destination table if process proc
        has been terminated successfully.
        """
        
                
        if proc.status=='done':
            self.del_proc(proc)

        if proc.status=='failed':
            self.del_proc(proc)
        
        
class Poll_dest(Proc):
    """Process that polls all destinations from a specified track management"""

    def init_proc(self,trackman_id):
        """
        Make an order to this trackman and ask for information about all
        available destination
        """
        #print 'Poll_dest',trackman_id
        self.make_order(trackman_id,'get_destinations')
        self.trackman_id=trackman_id # store management to deal with

    def put_info(self,info,sender_id=''):
        """
        Receive information about destinations from the trackman
        and pass it back to the task.
        """
        info['trackman_id']=self.trackman_id
        self.dest_ids=info['dest_ids']

        # update task with new destination from this track management
        self.task.put_info(info)
        
        #print 'put_info:dest_ids=',self.dest_ids
        #self.set_status('done')
