###############################################################################
# IMPORTS

from classlib.base import *
from RandomArray import *
import copy
###############################################################################
class Mod(Module):
    """
    Class, containing all data that are specific to a particular user
    (Data and methods that apply to all user modules of
    the same type are implemented in the respective user class
    object).
    """
    def __init__(self,obj,
                 pos=[0,0,0],
                 status='edit'):
        
        # back-link with simulation user object
        #self.obj=obj

       
        
        #  initialization, common to all modules
        self.init_common(obj)
        
        
        #self.modind=-1 # index in user array struct
        self.id='No id' # determined by add_mod of the simulation object

        # current status
        self.status=status

        #'Current position',pos
        #self.position=pos[0:2]
        #print 'init:self.position',self.position
        self.set_position(pos[0:2])

        
        
        # add this module to user simulation object
        self.obj.add_mod(self)

        

        # default management is empty, means not registered from the beginning
        # The registration will take place at latest
        # when user arrives at the first stop.
        # The default management is then self.obj.man_id
        self.man_id=''

                
        # current module that is dealing with this user
        self.location=None
        self.location_category=''
        
        # carrier for boarding and unboarding
        self.car=None

        # track elem to go to after boarding
        self.elem=None
        
        # initial destination id of user
        self.dest_id=''
        self.orig_id=''
        self.target_mod=None
        self.target_place=''

        # messgae stuff
        self.mes_ops['choose_dest']=self.choose_dest

        self.init_module_specific()

    def init_module_specific(self):
        """
        module specific init.
        """
        pass

    def init_costs(self):
        """
        Cost initialization of defaults
        """
        self.set_costs()
        self.reset_costs()
        
        
    def make_costspage(self,parent,widgets):
        """
        Default cost gui widget creation.
        Creates widgets in parent frame. All widgets are put in the
        widgets directory with a string as keyword. These widgets
        can be configured in the corresponding update method.

    
        """
        widgets['label']=Tkinter.Label(parent,
                                       text='No cost functions implemented.'+\
                                       '\nSee user management "'+self.man_id+\
                                       '"\n  for debits of this user')
        widgets['label'].pack(expand=1,fill='both')

    def update_costspage(self,panel,widgets):
        """
        Update of widgets on costs page. This method is called
        when cost-page needs to be updated.

        """
        pass
        
    def format_costs(self):
        """
        Return formatted costdata.
        """
        return [],{}
    
    def format_params(self):
        """
        Return formatted costdata.
        """
        descriptors=[
            {'key':'time_boarding',
             'name':'Boarding time',
             'type':'real',
             'value':{'min':0,'max':3600},
             'printformat':'%5.2f',
             'units':{'s':1},
             'width':1,
             'mode':'write',
             },
            {'key':'time_unboarding',
             'name':'Exit time',
             'type':'real',
             'value':{'min':0,'max':3600},
             'printformat':'%5.2f',
             'units':{'s':1},
             'width':1,
             'mode':'write',
             },
            {'key':'velocity_walk',
             'name':'Walking speed',
             'type':'real',
             'value':{'min':0,'max':3600},
             'printformat':'%5.2f',
             'units':{'m/s':1},
             'width':1,
             'mode':'write',
             },
            ]
        dataset={'velocity_walk':self.velocity_walk,
                 'time_boarding':self.time_boarding,
                 'time_unboarding':self.time_unboarding}

       

        return descriptors,dataset

    def make_panel(self,panel,**kw):
        """
        Make widgets.
        May be overwritten by specific class.
        """
        # initialize widgets with Non objects such that they are not updated
        # during the make phase.
        
        panel.make_standardpanel()
    

        # add a parameters page
        panel.make_page('Parameters',self.make_parampage,self.update_parampage)
        
        
        
    def make_parampage(self,parent,widgets):
        descriptors,dataset=self.format_params()
        widgets['param_data']=gui.DatasetEditor(descriptors,dataset,
                                               parent,
                                               applyentry_command=\
                                                   self.apply_param_gui,
                                               applybutton_command=\
                                                   self.apply_param_gui)
        widgets['param_data'].pack(expand=1,fill='both')
        
    def update_parampage(self,panel,widgets):
        """
        Update of widgets on costs page. This method is called
        when cost-page needs to be updated.

        """
        descriptor,dataset=self.format_params()
        widgets['param_data'].update_contents(dataset)

    def apply_param_gui(self,paramdatawidget,param_data):
        self.set_params(param_data)
            
        descriptor,dataset=self.format_params()
        paramdatawidget.update_contents(dataset)

    

    def reset_statistics_mod(self):
        """
        Reset or initialize statistical variables.
        """
        self.trips_total=0
        self.waitingtime_total=0.0
        self.time_start_waiting=0.0

        self.triptime_total=0.0
        self.time_start_traveling=0.0

    def format_statistics(self):
        """
        Format man specific contents for statistical data.
        """
        descriptor=[
            {'key':'name_orig',
             'name':'Name of current orig.',
             'type':'text',
             'width':1,
             'mode':'read',
             },
            {'key':'name_dest',
             'name':'Name of current dest.',
             'type':'text',
             'width':1,
             'mode':'read',
             },
            {'key':'trips_total',
             'name':'Total # of trips',
             'type':'integer',
             'printformat':'%9d',
             'valid':{'min':0,'max':10**8},
             'mode':'read',
             'width':0.2,
             },
            {'key':'triptime_total',
             'name':'Total time traveled',
             'type':'real',
             'printformat':'%8.2f',
             'valid':{'min':0,'max':10**8},
             'mode':'read',
             'width':0.4,
             'units':{'h':1}
            },
            {'key':'waitingtime_total',
             'name':'Total waiting time',
             'type':'real',
             'printformat':'%8.2f',
             'valid':{'min':0,'max':10**5},
             'mode':'read',
             'width':0.4,
             'units':{'h':1}
             },
            {'key':'triptime_average',
             'name':'Average trip time',
             'type':'real',
             'printformat':'%8.2f',
             'valid':{'min':0,'max':10**5},
             'mode':'read',
             'width':0.4,
             'units':{'s':1}
             },
            {'key':'waitingtime_average',
             'name':'Average waiting times',
             'type':'real',
             'printformat':'%8.2f',
             'valid':{'min':0,'max':10**5},
             'mode':'read',
             'width':0.4,
             'units':{'s':1}
             },
            ]
        
        if self.trips_total>0:
            triptime_average=self.triptime_total/self.trips_total
            waitingtime_average=self.waitingtime_total/self.trips_total
        else:
            triptime_average=0.0
            waitingtime_average=0.0
        if self.dest_id:
            name_dest=self.sim.get_mod(self.dest_id).name
        else:
            name_dest='None'


        if self.orig_id:
            name_orig=self.sim.get_mod(self.orig_id).name
        else:
            name_orig='None'
            
        #triptime_average
        dataset={'trips_total':self.trips_total,
                 'triptime_total':self.triptime_total/3600,
                 'waitingtime_total':self.waitingtime_total/3600,
                 'triptime_average':triptime_average,
                 'waitingtime_average':waitingtime_average,
                 'name_dest':name_dest,
                 'name_orig':name_orig,
                 }
    
        return descriptor,dataset

    def get_state_mod(self,state_image):
        '''
        This function removes all module-specific pointers from the
        state-image, see get_state function.
        To be overwritten by module-specific function.
        '''
        #if self.obj.state.db.has_key(self.no):
        #    if (self.car==None)&(self.elem==None):
        #        print 'WARNING in get_state_mod of',self.id
        #        print '  boarding/unboarding time=',self.obj.state[self.no]
        #        print '  but there is no car or elem to goto'
        #    else:
        #        print 'OK in get_state_mod of',self.id
        #        if self.car!=None:
        #            print '  found car',self.car.id
        #        if self.elem!=None:
        #            print '  found elem',self.elem.id
            
        if self.location!=None:
            state_image['location']=self.location.id
        if self.car!=None:
            state_image['car']=self.car.id
        if self.elem!=None:
            state_image['elem']=self.elem.id
        if self.target_mod!=None:
            state_image['target_mod']=self.target_mod.id

        self.get_state_mod_specific(state_image)
        
        return state_image

    def get_state_mod_specific(self,state_image):
        '''
        This function removes all module-specific pointers from the
        state-image, see get_state function.
        To be overwritten by module-specific function.
        '''
        pass
    
    def rebuild_pointers(self):
        """
        Convert ids to object pointers.
        """
        #if self.id=='base.user.test_driver/396':
        #    print 'rebuild_pointers  ',self.target_mod
        #    print '  location',self.location
        #    print '  car',self.car
        #    print '  elem',self.elem
            
       
        #replace id by pointer
        self.location=self.sim.rebuild_pointer(self.location)
        self.car=self.sim.rebuild_pointer(self.car)

        if hasattr(self,'elem'):
            self.elem=  self.sim.rebuild_pointer(self.elem)
        self.target_mod= self.sim.rebuild_pointer(self.target_mod)
       
       
    def set_state_mod(self):
        '''
        This function is called during the opening process of a simulation
        file. Here goes all the stuff that gets lost during the save
        process (essentially pointers to functions).  It is also called
        during the module initialization. 

        '''
        #if not self.sim.init_flag:
        #self.rebuild_pointers()

        #print 'set_state_mod',self.orig_id
        self.mes_ops['choose_dest']=self.choose_dest
        # obsolet...
        #self.mes_ops['set_dest']=self.set_dest
        #if string.atof(self.get_version())<0.2:
        if not hasattr(self,'elem'):    
            # track elem to go to after boarding
            self.elem=None

        if not hasattr(self,'car'):    
            # car where user is sitting
            self.car=None

        if not hasattr(self,'orig_id'):    
            # car where user is sitting
            self.orig_id=''
            
        
        

        # devide previous boarding+unboarding time by 2
        # to create new time_boarding and separated time_unboarding
        if not hasattr(self,'time_unboarding'):
                self.time_unboarding=self.time_boarding/2.0
                self.time_boarding=self.time_unboarding

        self.set_state_mod_specific()


    

    def set_state_mod_specific(self):
        '''
        This function is called during the opening process of a simulation
        file. Here goes all the stuff that gets lost during the save
        process (essentially pointers to functions).  It is also called
        during the module initialization. 

        '''
        pass
                
            
            
    def init_params(self,param={}):
        """
        Initialize parameters. Is called only once during init_common.
        Here all constant attributes should be defined.
        To oververwrite.
        """
         # set default parameters
        self.velocity_walk=3.0/3.6 # default walking speed in m/s
        self.time_boarding=5.0
        self.time_unboarding=5.0
        
    def set_params(self,params={}):
        """
        Set arbitrary parameters and make them effective, if possible.
        """
        #print 'set_params of ',self.id
        #self._params.update(params)
        self.time_boarding=params.get('time_boarding',\
                                      self.time_boarding)
        self.time_unboarding=params.get('time_unboarding',\
                                        self.time_unboarding)
        
        self.velocity_walk=params.get('velocity_walk',\
                                        self.velocity_walk)
      
        
    def set_position(self,position):
        """
        Setting current 2D position of user.

        """
        
        self.__pos__=position
        #print 'user.set_position=',self.__pos__
        
    def get_position(self):
        """
        Get current 2D position of user in array format.
        """
        #print 'user.get_position=',self.__pos__

        
            
        if self.obj.position.db.has_key(self.no):
            return self.obj.position[self.no][0].tolist()
        else:
            return self.__pos__
        
    def walkto(self,target_mod,target_place,args=None):
        """
        Make this user walk to certain target place within a
        target module.

        The exact coordinates of the target place
        is returned by the get_walkto_position method of the
        target_module.

        When the user arrived at the target place the
        terminate_walkto method  of the
        target_module.
        """
        self.target_mod=target_mod
        self.target_place=target_place
        
        #print self.id,'walkto',target_mod.id,target_place
        
        self.obj.position[self.no]=array(self.get_position())

        self.obj.position_target[self.no]=array(
            target_mod.get_walkto_position(self,args))
        

        # 2D vector across distance that needs to be walked
        walk_vec= self.obj.position_target[self.no][0]-\
                  self.obj.position[self.no][0]

        

        # compute normalized vector vector for one single step
        self.obj.step[self.no]=1/sqrt(dot(walk_vec,walk_vec))\
                                *self.velocity_walk*self.obj.iterate_interval\
                                *walk_vec

        #print '   self.get_position()=',self.get_position()
        #print '   walkto:walk_vec=',walk_vec
        #print '   step=',self.obj.step[self.no]
        
        self.obj.zerovec=zeros(len(self.obj.step.arrays))

        
        #print '   from ',self.obj.position[self.no][0],\
        #      ' to ',self.obj.position_target[self.no][0]
        #print '    len(self.obj.position.db)',len(self.obj.position.db)
        
        self.plot()
        
    def transfer(self,location,trans=identity(4)):
        '''
        General method to enter 
        User changes to a new location.
        Currently the new location
        must be a track element with the aprobriate attributes.
        '''

        # so let's first update x/y position from transformation
        self.set_position([trans[0,3],trans[1,3]])

        #print self.id,'transfers to',location.id,location.status
        #print 'transfers:self.position=',self.get_position()
        #if self.location:
        #    print 'from',self.location.id

        # OK, what kind of location is it ?
        category=self.obj.sim.split_id(location.id)['category']


        
        # can user enter this location ?
        # if user is still in edit mode it can only enter a track element
        if hasattr(location,'enter')&\
               (location.status not in ('passive','edit'))&\
               (not ( (self.status=='edit') & (category!='track') )):

            # make sure that user is active
            self.set_status('active')

            #print self.id,'transfers to',location.id,location.status
            # exit old location, if existent
            if self.location:
                #print '  self.location:',self.location.id
                if self.debug>=5:
                    print '***Transfer',self.id,\
                          '\n   from',self.location.id,'to',location.id
                    
                self.location.exit(self)
                # this is to trap a bug...
                #try:
                #    self.location.exit(self)
                #except:
                #    print 'Error transferring',self.id,\
                #          'from',self.location.id,'to',location.id
                #    return 1
            #else:
            #    print '  self.location:',self.location
                
            # enter new location
            location.enter(self)

            
            #print 'transferred',self.id,self.status,'to',location.id,category
            if  category=='track': # is it a track element i.e. stop

                if self.location_category=='car':
                    # last location has been in a carrier
                    # this means user just arrived from a trip
                    
                    # statistics
                    # compute traveling time
                    self.triptime_total+=self.obj.sim.time-\
                                         self.time_start_traveling
                    # increase number of trips by one
                    self.trips_total+=1
                elif self.location_category=='':
                    #print '  first location for',self.id
                    # Are we already inscribed in a user management ?
                    # ... in case it is out first trip
                    if self.man_id=='':
                        # let's inscribe in default  user management.
                        self.man_id=self.obj.man_id
                        self.send_mes({'to':self.man_id,
                                       'op':'register'
                                       })
                        # safe origin here for the first time
                        self.orig_id=location.id

                
                # we are at a stop, so let's
                # make a trip to a new destination, no matter what we have
                # done before

                

                # walk to a random place within the stop area
                # print self.id,'transfer to',location.id
                self.walkto(location,'random')
                
                
            elif category=='car':
                if self.location_category=='track':
                    # save stop as origin before entering car
                    self.orig_id=self.location.id
                    
                self.car=None
                
                
                # initialize traveling time
                self.time_start_traveling=self.obj.sim.time

                if self.obj.netmon: self.del_plot()

            else:
                return 1
            
            # store new location as current location
            self.location=location
            self.location_category=category
            return 0
        
        else:
            #print self.id,'transfer to',location.id,' failed for',self.id
            return 1
        
    def waiting_at(self,elem):
        """
        A track element put this user into waiting position.
        So do something!
        """
        #print self.id,'waiting_at',elem.id
        if self.location_category=='track':
            # Now send a message to the user management,
            # asking for available destinations
            self.send_mes({'to': self.man_id,
                           'op':'get_userdests'
                           })
            # user management will return destination and call the
            # choose_dest method
            
        else:
            print 'WARNING:',self.id,' is waiting.'
            print '  Do not know what to do in',\
                  self.location.id

    
        
    def exit_at(self,elem):
        """
        A track element tells user that he arrived at the exit.
        He may do  something, just as with the waiting_at call...
        """
        # or do nothing, which is the default
        pass
    
    def choose_dest(self,message):
        """
        Chooses a new destination for the user
        from the destinations returned by the user management.
        """
        n=len(message['dest_ids']) # number of of availlable destinations
        

        if n>0:
            # choose random destination
            
            self.dest_id=message['dest_ids'][randint(n)]
            #print self.id,'choose_dest',self.dest_id
            
        else:
            # quick and diry:
            self.dest_id=self.location.id # go back where we started 
            #self.dest_id=''
            print 'WARNING:',self.id,'no destinations in choose_dest()'
        self.start_trip()
    
    def start_trip(self):
        """
        Make a trip from current location, which is assumed to be a stop,
        to the stop dest_id...and queue up!
        """
        #print '>>>',self.id,'start_trip at',self.obj.sim.time
        # make an order to  the user management.
        self.send_mes({'to':self.man_id,
                       'op':'trans',
                       'dest_id':self.dest_id
                       })

        # statistics:
        # initialize waiting time here..when user made the decision where to go
        self.time_start_waiting=self.obj.sim.time
        
        #print self.id,'start_trip to',self.dest_id

        # walk to queue
        self.walkto(self.location,'queue_up')

        
    def start_boarding(self,car):
        '''
        User starts boarding carrier car.
        '''
        #print self.id,':start_boarding',car.id,'at',self.location.id
        if hasattr(car,'enter'):
            #if self.obj.netmon: self.del_plot()


            # statistics
            # computed waiting time here
            self.waitingtime_total+=self.obj.sim.time-\
                                     self.time_start_waiting
                
            self.car=car
            # add user into arrays to simulate boarding
            self.obj.add_user_wait_arrays(self,self.time_boarding)
            
            # terminate_boarding will be called after boarding simulation
            # has been terminated
            return 0
        else:
            return 1
            
    def terminate_boarding(self):
        '''
        User terminates boarding 
        '''
        #print self.id,':terminate_boarding to',self.car.id
        ans=self.transfer(self.car)
        self.car=None
        return ans

    
    def start_unloading(self,elem):
        '''
        User starts unboarding carrier car.
        '''
        #print self.id,'start_unloading:',car.id
        if hasattr(elem,'enter'):
            self.elem=elem
            # add user into arrays to simulate boarding
            self.obj.add_user_wait_arrays(self,self.time_unboarding)
            
            # terminate_unloading will be called after unboarding simulation
            # has been terminated
            return 0
        else:
            # can't enter this module
            return 1
            
    def terminate_unloading(self):
        '''
        User terminates unboarding. 
        '''
        #print self.id,':terminate_unboarding to',self.elem.id
        #
        # transfer user to the destination track element, using the
        # transform coordinates of the present location, the carrier
        ans=self.transfer(self.elem,self.location.get_transform())
        self.elem=None
        return ans

        
    def remove(self):
        '''
        Removal of user from location.
        '''
        if hasattr(self.location,'exit'):
            if self.location.exit(self):
                return 1 # cannot delete user from location
            else:
                # remove user from arrays
                self.obj.del_user_walkto_arrays([self])
                self.obj.del_user_wait_arrays([self])
                # unregister user from user management
                self.send_mes({'to':self.man_id,
                               'op':'unregister'
                               })
                return self.set_status('edit')

        else:
            return 1  # user cannot be removed from this particular location


    def set_status(self,newstatus):
        """
        Central function to change status of user.
        Here is the place where other simulation objects can
        be called when the status changes.
        """
        # currently we inform no management, the user is unsupervised!
        
        #print 'set_status:' ,self.status, ' of user'\
        #      , self.id,' to ',newstatus
        if self.status!=newstatus:
            pass
        
        self.status=newstatus
    
    
    #def update_man(self):
    #    '''
    #    This function is called while placing or after changes
    #    of destinations or status.
    #    '''

        # composing the user data that needs to be send to the user man
        # For safety reasons we do not send objects (or pointers in general),
        # just strings, numbers or ids (also strings).

     #   if self.location:
     #       location_id=self.location.id
     #   else:
     #       location_id=''
        

        # send status massage to user management
      #  self.send_mes({'to':self.man_id,
      #                 'op':'update_user',
      #                 'status':self.status,
      #                 'location_id':location_id,
      #                 })
        

    #def del_man(self):
    #    '''
    #    Function to remove this user from the user management database.
    #    This Function is called when user is going to be deleted
    #    from the network or changes of the user management.
    #    '''
    #    self.send_mes({'to':self.man_id,
    #                   'op':'del_user'})
    #    
    
        
    def print_state(self,debuglevel=0):
        if debuglevel<=self.debug:
            self.show_info()
            
    
        
    def plot(self,canvas=None):
        '''General plot function'''
        
        if not canvas: canvas=self.obj.netmon # =None if netmon not invoked

        if canvas:
            # first remove from canvas if existent
            tags_found=canvas.find_withtag(self.id)
            if tags_found:
                canvas.delete(self.id)

            x,y=self.get_position()
            d=self.obj.diameter_netmon
            canvas.create_oval(x-d/2,y-d/2,x+d/2,y+d/2,
                               outline=self.obj.color_netmon,
                               fill=self.obj.color_netmon_hair,
                               tag=self.id)
        
    def del_plot(self,canvas=None):
        if not(canvas): canvas=self.obj.netmon # =None if netmon not invoked
        if canvas:
            tags_found=canvas.find_withtag(self.id)
            if tags_found:
                canvas.delete(self.id)

    def show_info(self):
        """
        Show info about driver.
        """
        print '------------------------------------------------------------'
        print self.id,' Status:',self.status

        if self.location:
            print '  Location:',self.location.id,\
                  ', Category',self.location_category
        if self.dest_id:
            print '  Next destination:',self.dest_id,\
                  'named ',self.sim.get_mod(self.dest_id).name

            

        if self.orig_id:
            print '  Origin:',self.orig_id,\
                  'named ',self.sim.get_mod(self.orig_id).name

        #if self.obj.state.db.has_key(self.no):
        #   print '  Currently boarding'
        
        if self.target_mod:
            print '  Current or last walk to',self.target_mod.id,\
                  self.target_place
        if self.car:
            print '  sitting/boarding car',self.car.id

        if self.elem:
            print '  track elem to go to after boarding:',self.elem
            
        if self.obj.state.db.has_key(self.no):
            print '  boarding/unboarding time=',self.obj.state[self.no]
       
        self.show_info_specific()

    def show_info_specific(self):
        """
        Show mod specific info about driver.
        """
        pass
        
class User(Obj):

    def init_obj(self):
        self.name='Test driver'
        self.info='Generic test driver'+\
                   'who makes one random trip after the other'
        
        # basic parameters
        #self.velocity_walk=3.0/3.6 # default walking speed in m/s
        #self.boarding_time=10  #seconds
        
        
        # appearance on the Netmon 2D visulalizer
        self.diameter_netmon=1.5  # diameter in points of the circle
        self.color_netmon_hair='#cd661d' # chocolate3 color of the circle inner
        self.color_netmon='#8b4513' # chocolate4 color of the circle outline
        self.iterate_interval=1.0 # second update

        # default management simulation object
        self.man_id='base.userman.experim'
        
        # arrays
        self.position=Array_db()
        self.position_target=Array_db()
        self.step=Array_db()

        self.state=Array_db()

        # multiadd stuff
        self.multiadd_panel=None
        if self.sim.mode=='gui':
            self.multiadd_data={}

        
        
        # type of user module (Mod is the standard one)
        self.Mod=Mod

        # define initialy a None-module (here single user) to be edited 
        self.editmod=None
        self.init_type()
        

           
    def init_type(self):
        """User-Type specific initialization. To be overwritten"""
        pass

    def iterate(self):
        '''Iterates user actions'''

        if (len(self.position.db)>0):
            # walk a step with all users
            self.position.arrays=self.position.arrays+self.step.arrays

        
            #print 'user.position.arrays=\n',self.position.arrays
            #print 'user.position.db=\n',self.position.db.keys()
            # move users on netmon
            if self.netmon:
                for user_no in self.step.db.keys():
                    #print 'user.position for user',self.mods[user_no].id,\
                    #      '\n',self.position[user_no][0,0],\
                    #      self.position[user_no][0,1]
                    #print 'step',self.step[user_no]

                    self.netmon.move_on_canvas(self.mods[user_no].id,
                                               self.step[user_no][0,0],
                                               self.step[user_no][0,1])

            # terminate all users that reached the target
            self.terminated_ind=nonzero(
                greater(self.zerovec,
                        sum((self.position_target.arrays-
                             self.position.arrays)*self.step.arrays,1)
                        )
                )
        
            #print 'dot=\n', sum((self.position_target.arrays-
            #                 self.position.arrays)*self.step.arrays,1)
            if len(self.terminated_ind)>0:           
                users_terminated=[]
                for ind in self.terminated_ind:
                   #print self.id,'.iterate.......................'
                   #print '  ind=',ind
                   #print '  user_no=',self.step.get_key_with_index(ind)
                   #print '  step.db=',self.step.db
                   #print '  mods=',self.mods
                   user_no=self.step.get_key_with_index(ind)
                   user=self.mods[user_no]
                   users_terminated.append(user)
                   #print 'terminate walk for ',user.id
                   user.set_position(self.position[user_no][0].tolist())

                # delete users from array
                self.del_user_walkto_arrays(users_terminated)

                # now initialize successive user action
                # by telling the target module that the user arrived
                # the target module may then contact the users
                for user in users_terminated:
                    #user.set_position(self.position[user_no].to_list())
                    user.target_mod.terminate_walkto(user)
            
        if (len(self.state.db)>0):
            # count down boarding times 
            self.state.arrays=self.state.arrays-self.iterate_interval
            #print 'user.state.arrays',self.state.arrays
        
            # get array index of those users which terminated boarding
            inds_term=nonzero(greater(0,self.state.arrays))
            if len(inds_term)>0:
                users_term=[]
                for ind in inds_term:
                    # get user module
                    #user=self.mods[self.state.get_key_with_index(ind)]
                    #try:
                    user=self.mods[self.state.get_key_with_index(ind)]
                    #except:
                    #    print 'Index error!! with ',ind
                    #    print 'inds_term=',inds_term
                    #    print 'modno=',self.state.get_key_with_index(ind)
                    #    print 'state.arrays=\n',self.state.arrays
                    #    print 'state.db=\n',self.state.db
                    #    print 'Modules:\n',self.mods
                    if user.elem:
                        # terminate unboarding if the user has a
                        # track element where to go to
                        user.terminate_unloading()
                    elif user.car:
                        # terminate boardingif the user has a
                        # carrier where to go to
                        user.terminate_boarding()
                    else:
                        print 'WARNING in iterate',self.id
                        print '  user',user.id,'terminated (un)boarding'
                        print '  without track element or car to go'
                        user.show_info()
                    
                    users_term.append(user)

                # now users who terminated boarding can be deleted from arrays
                self.del_user_wait_arrays(users_term)

        # do Object specific iterations
        self.iterate_specific()

    def iterate_specific(self):
        """
        Additional Object specific iterations, to overwrite
        """
        pass
        
    def get_state(self):
        if self.netmon:
            self.netedit_gui.default()
            
        state_image={}
        
        # state_image['']=self.
        state_image['mods']=self.mods
        state_image['mod_no_current']=self.mod_no_current
        state_image['position']=self.position
        state_image['position_target']=self.position_target
        state_image['step']=self.step
        state_image['state']=self.state
       
        #state_image['']=self.

        self.get_state_specific(state_image)
        
        return state_image
        
   

    def set_state(self,state_image):
        #print 'set_state',self.id
        self.mods=state_image['mods']
        self.mod_no_current=state_image['mod_no_current']
        self.position=state_image['position']
        self.position_target=state_image['position_target']
        self.step=state_image['step']
        self.state=state_image['state']
        self.zerovec=zeros(len(self.step.arrays))
        
        #self.=state_image['']
        for mod in self.mods.values():
            mod.set_state(self)

        self.set_state_specific(state_image)
        
    def link(self):
        self.netmon=self.try_link_with('base.net.netmon')
        self.netedit_gui=self.try_link_with('base.net.netedit_gui')
        #print 'link',self.id,self.sim.init_flag
        if self.sim.init_flag:
            #print 'link',self.id,self.sim.init_flag
            
            if len(self.mods)>0:
                # print 'append_obj_iterate',self.id
                self.sim.append_obj_iterate(self)
                
            #if (len(self.step.db)>0)|(len(self.state.db)>0):
            #    # Append simulation object to maincycle
            #    self.sim.append_obj_iterate(self)

            
            

        self.link_specific()

    def link_specific(self):
        """
        Object specific linking.To be overwritten.
        """
        pass

    def rebuild_pointers(self):
        """
        Rebuild pointers that have been removed before saving.
        This function is called AFTER all objects have been
        initialized and cross-linked.
        """
        if self.sim.init_flag:
            for mod in self.mods.values():
                mod.rebuild_pointers()
                if self.netmon:
                    if mod.location_category!='car':
                        mod.plot()
    
    
    def add_gui(self):
        self.netmon.set_action('add user',
                               callback_click = self.add_click,
                               callback_motion = self.add_click,
                               callback_rightclick=self.add_end,
                               callback_termination = self.add_termination)
        
        
    def add_click(self,event):
        
        canvas=event.widget
        if canvas==self.netmon.canvas:
            self.event_past = event
            # unmagnify coords
            x,y=self.netmon.get_unmagnified_canvas_coords(event)

            self.editmod=self.Mod(self,pos=[x,y,0])
            self.editmod.plot()
            if self.debug>3:
                print 'Initialized module ' + self.editmod.id
            
            self.netmon.set_action('add user',
                                   callback_click = self.add_place,
                                   callback_motion = self.add_motion,
                                   callback_pressmotion = self.add_motion,
                                   callback_rightclick=self.add_end,
                                   callback_termination = self.add_termination)
            # this is a bit dirty at the moment...improve also at
            # add_motion
            event.x=canvas.canvasx(event.x)
            event.y=canvas.canvasy(event.y)
            self.event_past = event
            
    def add_motion(self,event):
        canvas=event.widget
        if canvas==self.netmon.canvas:
            if self.rescaled_flag:
                #print ' reset mod to current pointer position'
                x,y=self.netmon.get_unmagnified_canvas_coords(event)
                self.editmod.set_position([x,y])
                event.x,event.y=self.netmon.magnify_xy(x,y)
                self.event_past = event
                self.editmod.plot()
                self.rescaled_flag=0
            else:
                x=canvas.canvasx(event.x)
                y=canvas.canvasy(event.y)
                self.netmon.move([(self.editmod.id,
                                   self.event_past.x, x,
                                   self.event_past.y, y)],noscale=1
                                 )
                event.x=x
                event.y=y
                self.event_past=event
    
    def add_place(self,event):
        canvas=event.widget
        error=self.place_gui(canvas.canvasx(event.x),
                             canvas.canvasy(event.y))
        if not error:
            self.add_click(event)
        else:
            x,y=self.netmon.get_unmagnified_canvas_coords(event)
            self.editmod.set_position([x,y])
            event.x,event.y=self.netmon.magnify_xy(x,y)
            self.event_past = event
            self.editmod.plot()


    def add_end(self,event):
        self.add_termination(event)
        self.netedit_gui.default()
        
    def add_termination(self,event):
        if self.editmod:
            if self.debug>=5:
                print 'Terminate Add module '+self.editmod.id
            self.del_mod(self.editmod)
            
            # delete from monitor
            self.editmod.del_plot()
            self.editmod=None

    def multiadd_gui(self,mod):
        """
        Opens a panel to ask how many and how to add automatically
        multiple modules.
        """
        # print 'multiadd',self.id

        # this is an template module, to get default parameters from 
        self.editmod=mod
        
        if not self.multiadd_panel:
            self.multiadd_panel=gui.Panel(
                None,
                title=self.name+' - add multiple',
                make_cmd=self.make_multiadd_panel,
                refresh_cmd=self.refresh_multiadd_panel,
                apply_cmd=self.apply_multiadd_panel,
                close_cmd=self.close_multiadd_panel)
        else:     
            self.multiadd_panel.refresh()
        return 1
    
    def make_multiadd_panel(self,parent,widgets):
        """
        Create widgets for panel.
        """
        #descriptors,data=self.format_multiadd()
        widgets['label']=Tkinter.Label(parent,
                                       text='Add multiple users to'+\
                                       ' the availlable stops')
        
        descriptors=[
            {'key':'number',
             'name':'How many?',
             'printformat':'%d',
             'type':'integer',
             'width':1,
             'value':{'min':0,'max':10**7},
             'mode':'write',
             },
            {'key':'zone',
             'name':'Initial zone to go',
             'type':'text',
             'width':1,
             'mode':'write',
             },
            ]
        self.multiadd_data={'number':50,
                            'zone':'everywhere',
                            'time_boarding':15.0,}

        # get defaults from prototype
        descriptors_params,data_params=self.editmod.format_params()

        
        descriptors+=descriptors_params
        self.multiadd_data.update(data_params)
        #print self.multiadd_data
        
        
        widgets['multiadd_data']=gui.DatasetEditor(descriptors,
                                                  self.multiadd_data,parent)
        
        
        widgets['label'].pack(expand=1,fill='both')
        widgets['multiadd_data'].pack(expand=1,fill='both')
        

    def refresh_multiadd_panel(self,parent,widgets):
        """
        Refresh widgets for panel with simulation parameters.
        """
        widgets['multiadd_data'].update_contents(self.multiadd_data)
        widgets['multiadd_data'].set_size(1.57*350,350)
        
    def apply_multiadd_panel(self,parent,widgets):
        """
        Apply widgets values for panel with simulation parameters.
        """

        self.multiadd_data=widgets['multiadd_data'].get_contents()
        self.multi_add()
        return 1

    def close_multiadd_panel(self,parent,widgets):
        """
        Do something before simpanel is closed.
        Here the pointer to the panel is set to None.
        """
        self.editmod=None
        self.multiadd_panel=None

    
    
    def multi_add(self,**multiadd_data):
        """
        Add automatically multiple modules using self.multiadd_data.
        """
        if multiadd_data:
            self.multiadd_data=multiadd_data
            
        # get a list with all currently availlable stops which are destinations
        stops=[]
        for stop in self.sim.get_stops().values():
            # decide here whether stoip is suitable for carriers
            if (stop.status=='active')&(stop.zone!='maintenance')&\
                   ((stop.zone==self.multiadd_data['zone'])|\
                    (self.multiadd_data['zone']=='everywhere')):
                #print '>>>>>>>>>> found', stop.id
                stops.append(stop)



        #print 'multi_add:',self.multiadd_data['number'],self.id
        if self.multiadd_data['number']>0:
            # make sure that this sim object is registered in maincycle now
            self.sim.append_obj_iterate(self)

        # compute number of users to put per stop
        if len(stops)>0:
            users_per_stop=self.multiadd_data['number']/int(len(stops))
        else:
            users_per_stop=0
            
        for stop in stops:
            for n in range(users_per_stop):
                user=self.Mod(self)
                user.set_params(self.multiadd_data)

                # compute initial coordinates
                # keep them at least 30m away from platform because
                # they have to inscribe first
                width_platform=2.0
                x,y=stop.berthssec.transform_polyline(\
                    [uniform(-0.8,1.0),\
                     uniform(-5.0*width_platform*stop.side,
                             -10*width_platform*stop.side)])
                z=0
                #stop.berthssec.length
                user.transfer(stop,array([[1,0,0,x],
                                          [0,1,0,y],
                                          [0,0,1,z],
                                          [0,0,0,1]]))
                
                
                
                                
    def place_gui(self,x,y,z=0):
        self.editmod.del_plot()

        # get track element module on which user has been placed
        obj,location_tag=self.netmon.find_obj_at_coords(x,y)
        if location_tag!='':
            location=self.sim.split_id(location_tag)['mod']

            # now transfer user to location but at coordinates
            # with 100% magnification
            x,y=self.netmon.unmagnify_xy(x,y)

            #print 'place_gui: make sure that users is in obj_iterate list'
            self.sim.append_obj_iterate(self)
            
            if location:
                # try to place user on location
                error= self.editmod.transfer(location,array([[1,0,0,x],
                                                             [0,1,0,y],
                                                             [0,0,1,z],
                                                             [0,0,0,1]]))
                if error:
                    self.editmod.plot()
                    return 1
                else:
                    return 0
                    
            else:
                self.editmod.plot()
                return 1
            
        else:
            self.editmod.plot()
            return 1
        
            
    def del_gui(self,event,tag):
        # get module with user out of tag string
        self.editmod=self.sim.split_id(tag)['mod']

        # delete module from data structures
        if not self.del_mod(self.editmod):
            # delete from monitor
            self.editmod.del_plot()
            self.editmod=None

        self.netedit_gui.del_mod()

    def show_info(self,event,tag):
        # does currently not work, because user has no explicit canvas object
        
        # get module with user out of tag string
        user=self.sim.split_id(tags)['mod']
        user.print_state()
        self.netedit_gui.show_info()
        
    def add_mod(self,mod):
        """
        Insert new module in the module database mods of the simulation object
        Initial module values are already initialized by the  module mod.
        """
        
        self.mod_no_current=self.mod_no_current+1
        mod.no=self.mod_no_current
        mod.id=self.get_mod_id(mod)
            
        # add module to modules database
        self.mods[self.mod_no_current]=mod

        #print 'add_mod: added mod',mod.id

        # make sure that this sim object is registered in maincycle now
        #self.sim.append_obj_iterate(mod.obj)
        
        return 0
    
    
    def del_mod(self,mod):
        """
        Delete module
        """
        # to remove the user, it must first be in edit status
        if mod.status!='edit':
            mod.remove() #  remove user
            # now, user is in edit status and can be deleted

        # now delete user instance when edit status
        if mod.status=='edit':
            # call specific delete function before removing the module
            self.del_mod_specific(mod)

            del self.mods[mod.no] # delete from user obj's database

            #if there are no more users to simulate,
            # 'remove user from maincycle:',self.type
            if len(self.mods)==0:
                self.sim.remove_obj(self)
                #print 'remove User object from maincycle:',self.type
                #self.sim.remove_obj_iterate(self)
            #    self.sim.remove_obj_output(self)
                
            return 0

        else: return 1 

    def del_mod_specific(self,mod):
        """
        Object specific Delete module. To be overwritten.
        """
        pass
        
            
    def add_user_wait_arrays(self,user,time):
        '''
        Add user to array structure.
        This is currently done when user starts waiting.
        '''
        self.state[user.no]=time
        #if self.mods:
        #    # make sure that this sim object is registered in maincycle now
        #    self.sim.append_obj_iterate(self)
   

        
    def del_user_walkto_arrays(self,users):
        """
        Delete list with users from arrays concerned with walkto methods.
        """
        for user in users:
            if self.step.db.has_key(user.no):
                del self.position[user.no]
                del self.position_target[user.no]
                del self.step[user.no]

        self.zerovec=zeros(len(self.step.arrays))

   
        
    def del_user_wait_arrays(self,users):
        '''
        Delete list of users from  array structure, used for boarding
        and unboarding.
        This is currently done after waiting.
        '''
        for user in users:
            if self.state.db.has_key(user.no):
                del self.state[user.no]

        # remove sim object from maincycle if there are no more modules
        #if len(self.mods)==0:
        #    self.sim.remove_obj_iterate(self)

    def report_statistics(self):
        """
        
        Return descriptor and dataset contents for statistical data
        about the target object.

        This overwrites the default method in obj.
        Takes special care of averaging waiting and trip-times.
        """
        #print 'report_statistics',self.id

        descriptor=[]
        dataset={}
        dataset_counter={}
        dataset_sum={}
        if self.multimod_flag:
            descriptor.append({'name':'Number',
                               'key':'number',
                               'type':'integer',
                               'printformat':'%9d',
                               'width':1,
                               'valid':{'max':10**8},
                               'mode':'read'},
                              )
            dataset['number']=len(self.mods)
            
            # cheap trick: ask modules for data and return average values
            if len(self.mods)>0:
                # init a conter and sum for reach key in the dataset
                nt=0
                for mod in self.mods.values():
                    # count users who made a trip
                    if mod.trips_total>0:nt+=1

                    self.cumulate_mod_results(mod.format_statistics,descriptor,
                                              dataset_sum,dataset_counter)

                #print '  Users who made a trip:',nt
                #print '  sum of waiting times:',dataset_sum['waitingtime_average']
                # average all data
                for key in dataset_sum.keys():
                    if nt>0: #key in ('waitingtime_average',
                        dataset[key]=dataset_sum[key]/nt
                    else:
                        #print '  ',key,dataset_sum[key]/dataset_counter[key]
                        dataset[key]=0.0
                        
        else:
            descriptor,dataset=self.format_statistics()
            
        self.report_statistics_specific(descriptor,dataset,
                                        dataset_sum,dataset_counter)
            
        return descriptor,dataset
