###############################################################################
# IMPORTS
from classlib.base import *
import Pmw, copy
import Tkinter,tkMessageBox,tkFileDialog,os,string
###############################################################################


class Netmon(Obj):
    """
    2D Net monitor.
    """
    
    def init_obj(self):
        
        self.color_background='beige'
        self.scrollsteps=20
        self.gui_flag=1  # initialize ONLY in gui simulation mode
        self.output_interval=0.2 # seconds
        if (not self.sim.objs.has_key(self.type))&\
           (self.sim.mode=='gui'):

            self.canvas_scale=1.0
            self.canvas_origin_x=0.0
            self.canvas_origin_y=0.0

            # background map stuff
            self.map_n_current=0
            self.editmap={}
            self.maps={}
            
            self.car_ids=[]
            
            # Create the ScrolledCanvas.
            self.scrolledcanvas = Pmw.ScrolledCanvas(
                self.sim.simwin,
                borderframe = 0,
                scrollmargin =0,
                canvasmargin = 10,
                #labelpos = 'n',
                #label_text = None,
                usehullsize = 1,
                hull_width = 1.6*300,
                hull_height = 300,
                )
            self.scrolledcanvas.pack(side='top',
                         padx = 5, pady = 5, fill = 'both', expand = 1)

            self.canvas=self.scrolledcanvas.component('canvas')
            self.canvas.configure(bg=self.color_background)
            #print self.canvas.keys()
            
            self.area=[0,0,500,400]
            self.canvas.create_rectangle(self.area,width=0,tag='area')
            self.scrolledcanvas.resizescrollregion()

            # default function  that is called when
            # no other action is specified
            self.function_default=None
            
            # key bindings
            self.terminate_action_callback=lambda event:0
            self.keymap={}
            self.keymap_action={}
            self.canvas.bind('<KeyPress>', self.keypress)
            self.canvas.bind('<Escape>', self.escape_action)
            
            self.canvas.bind('<Left>', self.scroll_left)
            self.canvas.bind('<Right>', self.scroll_right)
            self.canvas.bind('<Up>', self.scroll_up)
            self.canvas.bind('<Down>', self.scroll_down)
            self.canvas.bind('<Prior>', self.zoom_out)
            self.canvas.bind('<Next>', self.zoom_in)
            
            # simulate a setp when pressing enter
            self.canvas.bind('<Return>',self.sim.sim_step_gui)
            
            self.action=None
            self.set_action('default')


            
            # fill view menue
            self.sim.menubar.addmenuitem('View', 'command',\
                                             'zoom out',
                                             command = self.zoom_out,
                                             label = 'Zoom out')
                
            self.sim.menubar.addmenuitem('View', 'command',\
                                         'zoom in',
                                         command = self.zoom_in,
                                         label = 'Zoom in')

            self.sim.menubar.addmenuitem('View', 'separator')
            self.sim.menubar.addmenuitem('View', 'command',\
                                         'zoom to 50% magnification',
                                         command = self.zoom_50,
                                         label = 'Zoom 50%')
            
            self.sim.menubar.addmenuitem('View', 'command',\
                                         'zoom to 100% magnification',
                                         command = self.zoom_100,
                                         label = 'Zoom 100%')
            
            self.sim.menubar.addmenuitem('View', 'command',\
                                         'zoom to 200% magnification',
                                         command = self.zoom_200,
                                         label = 'Zoom 200%')

            
                
            self.sim.menubar.addmenuitem('View', 'separator')

            # Initialise the checkbutton for show_ghosts to 0:
            self.var_show_ghosts = Tkinter.IntVar()
            self.var_show_ghosts.set(0)
            
            self.sim.menubar.addmenuitem('View', 'checkbutton',
                                         'Indicate ghost vehicles on/off.'+\
                                         '\n Mainly for debugging'+\
                                         ' merge/demerge algorithms',
                                         label = 'Indicate ghosts',
                                         command = self._toggle_show_ghosts,
                                         variable = self.var_show_ghosts)
            
            self.sim.menubar.addcascademenu('Edit', 'Map',
                                            'Edit background map',
                                            #traverseSpec = 'z',
                                            tearoff = 1)
            
            self.sim.menubar.addmenuitem('Map', 'command',\
                        'Add a background map to the 2D network monitor',
                                         command = self.open_map,
                                         label = 'Add ...')

            self.sim.menubar.addmenuitem('Map', 'command',\
                        'Move background maps on 2D network monitor',
                                         command = self.move_map_gui,
                                         label = 'Move')

            self.sim.menubar.addmenuitem('Map', 'command',\
                        'Delete background maps from 2D network monitor',
                                         command = self.del_map_gui,
                                         label = 'Delete')


            self.sim.menubar.addmenuitem('Tools', 'command',\
                                     'Save graphics as Postscript...',
                                     command = self.saveas_ps,
                                     label = 'Save graphics as Postscript...')

            self.var_make_snapshots = Tkinter.IntVar()
            self.var_make_snapshots.set(0)
           
            self.sim.menubar.addmenuitem('Tools', 'checkbutton',
                                         'Start recording a picture sequence'+\
                                         '\n in Postscript format '+\
                                         ' in current directory',
                                         label = 'Make PS snapshots',
                                         command = self._toggle_make_snapshots,
                                         variable = self.var_make_snapshots)

            # counter for naming files with snapshots
            self.shot_count=0
            
    def get_state(self):
        '''Prepare state image for save operation'''

        # do the following  actions only in gui mode

        if self.sim.mode=='gui':
            # delete bitmaps to safe memory
            maps_save={}
            for map in self.maps.values():
                # safe all features of map except for the bitmaps
                maps_save[map['tag']]={}
                for key in map.keys():
                    if key!='bitmaps':
                        maps_save[map['tag']][key]=map[key]
        else:
            # recover map info for non-gui-mode
            maps_save=self.maps
        
        return {'color_background':self.color_background,
                'area':self.area,
                'maps':maps_save,
                'map_n_current':self.map_n_current}

    def set_state(self,state_image):
        # read state
        #print 'objlib/base/net/netmon.py.set_state'
        self.map_n_current=state_image.get('map_n_current',0)
        self.color_background=state_image.get('color_background','beige')
        self.area=state_image.get('area',[0,0,500,300])

        # do the following init actions only in gui mode
        if self.sim.mode=='gui':
            self.canvas.configure(bg=self.color_background)
            self.set_area()

            # generate bitmaps for backaground maps, if not already present
            if not self.maps:
                for map_save in state_image.get('maps',{}).values():
                    map=self.create_map(template=map_save)
                    if map:
                        self.add_map(map)
                    else:
                        #print 'Could not add map ',map_save['file_name']
                        pass
        else:
            # simulation is in non-gui mode
            # this is just to store the map info somewhere till it gets saved
            self.maps=state_image['maps']
                
    def close(self):
        """
        Does necessary action before this simulation object is
        deleted from memory.

        It can be assumed that the state of the object has been
        saved. So what remains is to clear up other things, essentially
        GUI stuff.
        """
        self.scrolledcanvas.destroy()
        self.sim.menubar.deletemenu('Map')
        self.sim.menubar.deletemenu('View')


    def saveas_ps(self):
        """
        Save graphics as Postscript...
        """
        filename =tkFileDialog.SaveAs(initialdir=self.sim.simdir,
                                      filetypes=[('Text files','*.ps'),
                                                 ('All files','*')]).show()
        bbox=self.canvas.bbox('all')

        #self.canvas.postscript(file=filename,
        #                       x=bbox[0],y=bbox[1],
        #                       width=2*bbox[2],height=2*bbox[3])
        self.canvas.postscript(file=filename)
        
        
        
    def zoom_out(self,event=None):
        """
        Zoom out of 2D canvas.
        """
        self.magnify(zoom=50)

    def zoom_in(self,event=None):
        """
        Zoom in  2D canvas.
        """
        self.magnify(zoom=200)

    def zoom_100(self,event=None):
        """
        Zoom to 100% magnification. 
        """
        self.magnify(mag=100)

    def zoom_200(self,event=None):
        """
        Zoom to 200% magnification. 
        """
        self.magnify(mag=200)

    def zoom_50(self,event=None):
        """
        Zoom to 50% magnification. 
        """
        self.magnify(mag=50)
        
    def _toggle_show_ghosts(self):
        """
        Called when show_ghosts buttom on toggle menue has been selected.
        """
        #print '_toggle_show_ghosts',self.var_show_ghosts.get()
        if not self.var_show_ghosts.get():
            # delete ghost pointers from canvas
            self.canvas.delete('ghosts_pointer')
            
    def _toggle_make_snapshots(self):
        """
        Called when make_snapshots  on toggle menue has been selected.
        """
        if not self.var_make_snapshots.get():
            self.shot_count=0
            
        #print '_toggle_show_ghosts',self.var_show_ghosts.get()
        #if not self.var_make_snapshots.get():
        #    # delete ghost pointers from canvas
        #    self.canvas.delete('ghosts_pointer')

    
    def open_map(self):
        """
        Open file-browser for background map
        """
                
        file_path=tkFileDialog.askopenfilename(
            initialdir=self.sim.simdir,
            filetypes=[('gif','*.gif'),('gif','*.GIF'),
                       ('ppm','*.ppm'),('ppm','*.PPM'),
                       ('All files','*')])

        if file_path:
            #print 'open_map: file_path',file_path
            #print '  ',string.split(os.path.dirname(file_path),'/')
            #print '  ',string.split(self.sim.simdir,os.sep)
          
            if string.split(os.path.dirname(file_path),'/')==\
               string.split(self.sim.simdir,os.sep):
                self.editmap=self.create_map(os.path.basename(file_path))
                if self.editmap:
                    #print 'created map:',self.editmap['tag']
                    self.plot_map(self.editmap)
                    self.set_action('add map',
                                    callback_click = self.add_map_click,
                                    callback_motion = self.add_map_click,
                                    callback_rightclick=\
                                                    self.add_map_termination,
                                    callback_termination = \
                                                    self.add_map_termination)
            else:
                if tkMessageBox.askyesno('Incorrect Path',
                                         'Sorry, the map needs to be\n'+\
                                         'in the same directory as the\n'+\
                                         'simulation file\n'+\
                                         'Please move your map file to:\n'+\
                                         self.sim.simdir+'\n\n'+\
                                         'Try again now?'):
                    self.open_map()
                    
            
    def add_map_click(self,event):
        """
        Plots map on canvas and attaches it to pointer.
        """
        #print 'add_map_click'
        canvas=event.widget
        if canvas==self.canvas:
            #x,y=self.get_unmagnified_canvas_coords(event)

            # this is a bit dirty at the moment...improve also at
            # add_motion
            event.x=canvas.canvasx(event.x)
            event.y=canvas.canvasy(event.y)
            self.event_past = event
            
            self.plot_map(self.editmap,(event.x,event.y))
            self.scrolledcanvas.resizescrollregion()
            self.set_action('add map',
                            callback_click = self.add_map_place,
                            callback_motion = self.add_map_motion,
                            callback_pressmotion = self.add_map_motion,
                            callback_rightclick=self.add_map_termination,
                            callback_termination = self.add_map_termination,
                            cursor='fleur')
            return 0
        return 1
    
    def add_map_place(self,event):
        """
        Detatch map from pointer and set unmagnified map coordinates.
        """
        #print 'add_map_place'
        canvas=event.widget
        if canvas==self.canvas:
            offset_x,offset_y=self.get_unmagnified_canvas_coords(event)
            self.editmap['offset_x']=offset_x
            self.editmap['offset_y']=offset_y
            self.add_map(self.editmap)
        self.editmap=None
        self.move_map_gui()

    
            
    def add_map_motion(self,event):
        """
        Move map with pointer.
        """
        #print 'add_map_motion',event.x,event.y
    
        canvas=event.widget
        if canvas==self.canvas:
            if self.rescaled_flag:
                # reset root to current pointer position 
                x,y=self.get_unmagnified_canvas_coords(event)
                event.x,event.y=self.magnify_xy(x,y)
                self.event_past = event
                #self.plot_map(self.editmap,(event.x,event.y))
                self.rescaled_flag=0
                return 0
            else:
                x=canvas.canvasx(event.x)
                y=canvas.canvasy(event.y)
                #print 'add_map_motion:',self.editmap['tag']
                #print '  from,to',self.event_past.x,self.event_past.y, x,y
                if len(canvas.find_withtag(self.editmap['tag']))==1:
                    self.move([(self.editmap['tag'],
                                self.event_past.x, x,
                                self.event_past.y, y)],noscale=1
                              )
                event.x=x
                event.y=y
                self.event_past=event
                return 0
        return 1    
    
    def add_map_termination(self,event):
        """
        Exit the add maps mode.
        """
        #print 'add_map_termination'
        if self.editmap:
            tag=self.editmap['tag']
            # is map an object on canvas?
            if len(self.canvas.find_withtag(tag))==1:
                self.canvas.delete(tag)
            self.editmap=None
            self.netedit_gui.default()

        return 0
            
    def del_map(self,map):
        """
        Remove background map
        from maps database of the 2D network monitor
        """
        #print 'del_map'
        tag=map['tag']
        if self.maps.has_key(tag):
            del self.maps[tag]
            
        # is map an object on canvas?
        if len(self.canvas.find_withtag(tag))==1:
            self.canvas.delete(tag)
        
    def create_map(self,file_name='',template={}):
        """
        Generates bitmap from file and returns a directory with
        data of background map.
        """
        if template:
             tag=template['tag']
             file_name=template['file_name']
             offset_x=template['offset_x']
             offset_y=template['offset_y']
        else:
            tag=self.sim.get_id(self.id,self.map_n_current)
            offset_x=0.0
            offset_y=0.0
            
        
        file_path=os.path.join(self.sim.simdir,file_name)
        #print 'create_map',file_name#,file_path,tag
        if os.path.isfile(file_path):
            # print 'construct bitmaps with info from file'
            map={'tag':tag,
                 'file_name':file_name,
                 'offset_x':offset_x,
                 'offset_y':offset_y}
            # bit maps are stored in dictionaries with magnification
            # as key. 
            map['bitmaps']={100:gui.PhotoImage(file=file_path)}
            return map
        else:
            return None

    def add_map(self,map,offset_x=0,offset_y=0):
        """
        Add map to the maps database of the 2D network monitor
        """
        #print 'add_map',map['tag']
        tag=map['tag']

        # make sure bitmaps are copied and not linked
        self.maps[tag]=map.copy()
        
        self.plot_map(self.maps[map['tag']])
        self.map_n_current+=1
        

    def plot_map(self,map,coords=()):
        """
        (Re)plot map at canvas position coords=(x,y).
        The magnification is always determined by the current canvas_scale.
        If no coordinates are provided, the current position is preserved
        in the current scale.
        """
        tag=map['tag']
        
        if  len(coords)!=2:
            x,y=self.magnify_xy(map['offset_x'],map['offset_y'])
        else:
            x,y=coords

        # check whether there is a bitmap for the current scale.
        # The bitmaps for different scales are stored in a dictionary
        # with mag as key, which is the integer of scale in per cent.
        mag=int(self.canvas_scale*100)
        
        if (not map['bitmaps'].has_key(mag))&(mag!=100):
            map['bitmaps'][mag]=map['bitmaps'][100].magnify(mag)
            # generate bitmap...currently *very* inefficient algorithm
            # see gui.PhotoImage.rescale()
            #self.sim.messagebar.message('state',
            #                            'Rescale map to '+`mag`+'% ...')
            #map['bitmaps'][mag]=gui.PhotoImage()
            #map['bitmaps'][mag].put(map['bitmaps'][100].rescale(mag))
            
            
            #map['bitmaps'][mag].put(
            #    map['bitmaps'][100].rescale(self.canvas_scale))

       
        
        # is map already an object on canvas?
        if len(self.canvas.find_withtag(tag))==1:
            #print 'configure existing map',tag,x,y
            self.canvas.coords(tag,x,y)
            self.canvas.itemconfig(tag,image=map['bitmaps'][mag])
            
        else:
            #print 'create_image',tag,x,y
            self.canvas.create_image(
                x,y,image=map['bitmaps'][mag],anchor='nw',
                tag=tag)
            
        self.adjust_area(tag)
        self.lower_maps()
        
    def move_map_gui(self,event=None):
        """
        Starts move mode for background maps.
        """
        self.set_action('move map',
                        callback_click = self.move_map_click,
                        cursor='fleur')
        
    def move_map_click(self,event):
        """
        Attach map to pointer.
        """
        canvas=event.widget
        if canvas==self.canvas:
            self.editmap=self.find_map(event)
            if self.editmap:
                # get unmagnified pointer position
                x,y=self.get_unmagnified_canvas_coords(event)
                
                # store absolute canvas coords
                event.x=canvas.canvasx(event.x)
                event.y=canvas.canvasy(event.y)
                self.event_past = event

                # determine offset between pointer and NW corner of map.
                # this will be required to place the map correctly
                
                self.editmap['pointeroffset_x']=(self.editmap['offset_x']-x)
                self.editmap['pointeroffset_y']=(self.editmap['offset_y']-y)
            
                self.set_action('move map',
                                callback_click = self.move_map_place,
                                callback_motion = self.add_map_motion,
                                callback_pressmotion = self.add_map_motion,
                                callback_rightclick=self.add_map_termination,
                                callback_termination =self.add_map_termination,
                                cursor='fleur')

    def move_map_place(self,event):
        """
        Place editmap at the coordinates of the event. 
        """
        #print 'move_map_place'
        canvas=event.widget
        if canvas==self.canvas:
            offset_x,offset_y=self.get_unmagnified_canvas_coords(event)
            self.editmap['offset_x']=offset_x+self.editmap['pointeroffset_x']
            self.editmap['offset_y']=offset_y+self.editmap['pointeroffset_y']
            del self.editmap['pointeroffset_x']
            del self.editmap['pointeroffset_y']
            self.plot_map(self.editmap)

        self.editmap=None
        self.move_map_gui()
        
    def del_map_gui(self,event=None):
        """
        Starts delete mode for background maps.
        """
        self.set_action('move map',
                        callback_click = self.del_map_click,
                        cursor='pirate')

    def del_map_click(self,event):
        """
        Delete map under pointer.
        """
        canvas=event.widget
        if canvas==self.canvas:
            self.editmap=self.find_map(event)
            if self.editmap:
                self.del_map(self.editmap)
                self.editmap=None
        self.del_map_gui()
        
  
    def lower_maps(self,map={}):
        """
        Put background maps to background on canvas.
        """
        if map:
            self.canvas.lower(map['tag'])
        else:
            for map in self.maps.values():
                self.canvas.lower(map['tag'])

    def set_defaultaction(self,function):
        """
        Default function that is called when ESC is pressed.
        """
        self.action_default=function
            
    def set_action(self,netmon_action,
                   callback_click=lambda event:0,
                   callback_doubleclick=lambda event:0,
                   callback_button2click=lambda event:0,
                   callback_motion=lambda event:0,
                   callback_pressmotion=lambda event:0,
                   callback_rightclick=lambda event:0,
                   callback_keymap={},
                   callback_termination=None,
                   cursor='', # for cursor names do man -K hand2 tk
                   ):

        if (self.action != netmon_action)|(netmon_action=='default'):
            
            #print 'terminate action is......',self.terminate_action_callback
            #print self.action,netmon_action
            termination_failed=self.terminate_action_callback(None)
            self.sim.messagebar.message('state','')
            self.canvas.configure(cursor='')

            
        else:
            termination_failed=0
                
        if not termination_failed:
            # define new callbacks for mouse and keyboard
            
            if self.debug>=5:
                print 'Netmon: terminate action "',self.action,\
                      '". Start with action "',netmon_action,'"'
                
                
            if netmon_action!='default':
                self.sim.messagebar.message('state','Action: '+netmon_action)
            self.canvas.configure(cursor=cursor)
                
            self.action=netmon_action
            self.canvas.bind('<ButtonPress-1>', callback_click)
            self.canvas.bind('<Double-Button-1>', callback_doubleclick)
            self.canvas.bind('<ButtonPress-2>', callback_button2click)
            self.canvas.bind('<Motion>',callback_motion)
            self.canvas.bind('<B1-Motion>',callback_pressmotion)
            self.canvas.bind('<ButtonPress-3>', callback_rightclick)


            if callback_termination:
                self.terminate_action_callback=callback_termination
            else:
                self.terminate_action_callback=lambda event=None:0
            
            if callback_keymap:
                self.keymap_action=callback_keymap

            self.canvas.focus_force()
            return 0
           
        else:
            #print 'Action "',self.action,'" not terminated.'
            return 1

    def escape_action(self,event=None):
        """
        Do standard termination of action if nothing else is specified. 
        """
        #print 'escape_action'
        if self.terminate_action_callback:
            self.terminate_action_callback(event)

        self.set_action('default')
        if self.action_default:
            self.action_default()

        
    def keypress(self,event):
        
        #print 'keypress, keymap_action',event.char
        if self.keymap_action.has_key(event.char):
            #print 'keypress, keymap_action',event.char
            self.keymap_action[event.char](event)

        elif self.keymap.has_key(event.char):
            #print 'keypress, keymap_action',event.char
            self.keymap[event.char](event)    

    def scroll_left(self,event):
        self.canvas.xview('scroll',-1,'units')

    def scroll_right(self,event):
        self.canvas.xview('scroll',+1,'units')

    def scroll_up(self,event):
        self.canvas.yview('scroll',-1,'units')

    def scroll_down(self,event):
        self.canvas.yview('scroll',+1,'units')

    def update_cars(self):
        """
        To be called when new cars are added to netsim or car
        parameters change.
        """
        # array
        car_nr=len(self.netsim.velocity.arrays)
        self.car_length_netmon=zeros((car_nr,1))
        
        #self.car_inds=zeros(self.car_length_netmon.shape)
        self.car_inds=reshape(arange(car_nr),(car_nr,1))
        self.car_ids=[]
        for car in self.netsim.cars.objs:
            #put(self.car_inds,(car.ind,0),car.ind)
            put(self.car_length_netmon,(car.ind,0),car.length_netmon)
            self.car_ids.append(car.id)
            


    def link(self):
        self.netsim=self.try_link_with('base.net.netsim')
        self.netedit_gui=self.try_link_with('base.net.netedit_gui')
        if self.netsim:
            self.sim.append_obj_output(self)
            self.update_cars()
            
    def output(self):
        # move all carriers in netsim
        if self.netsim.cars.db:
            self.cartrans=self.netsim.segs.take(
                self.netsim.segs_ind.arrays+\
                (self.netsim.path_sec/self.netsim.length.arrays *
                 self.netsim.segs_total.arrays).astype(Int)
                )
        
            self.offset=self.cartrans[:,0:2,3]
            self.mult=self.cartrans[:,0:2,0]
            self.path_seg=reshape(self.netsim.path_sec %\
                           self.netsim.seg_length.arrays,(len(self.mult),1))
        
            # compute nose and tail positions based on Netsim's
            # offset and xy multipliers
            self.pos=concatenate((self.offset+self.mult*self.path_seg,
                                  self.offset+self.mult*
                                  (self.path_seg-self.car_length_netmon),
                                  self.car_inds
                                  ),(1)).tolist()
            if self.var_show_ghosts.get():
                self.canvas.delete('ghosts_pointer')
                map(self.move_car_show_ghost,self.pos)
                self.scale_obj('ghosts_pointer')
            else:
                map(self.move_car,self.pos)
                
            self.scale_obj('car')

        if self.var_make_snapshots.get():
            #print 'snapshotsget()',self.canvas.bbox('all')
            bbox=self.canvas.bbox('all')
            self.shot_count+=1
            #print 'save to',self.sim.simdir+os.sep+'shot'+\
            #           '%03d'%(self.shot_count)+'.ps'
            #print 'save to',self.sim.simdir+os.sep+\
            #os.path.splitext(os.path.basename(self.sim.simfile_name))[0]+\
            #                       '_shot'+'%03d'%(self.shot_count)+'.ps'
            #print 'whatis',os.path.splitext(self.sim.simfile_name)[0]+\
            #                       '_shot'+'%03d'%(self.shot_count)+'.ps'
            #print 'simfile_name=',self.sim.simfile_name
            self.canvas.postscript(file=self.sim.simdir+os.sep+\
                os.path.splitext(os.path.basename(self.sim.simfile_name))[0]+\
                                   '_shot'+'%03d'%(self.shot_count)+'.ps',
                                   x=bbox[0],y=bbox[1],
                                   width=2*bbox[2],height=2*bbox[3])
                # % gifsicle --delay=10 --optimize --loop *.gif > anim.gif
            
    def move_car(self,pos):
        x_tail,y_tail,x_nose,y_nose,car_ind=pos
        self.canvas.coords(self.car_ids[int(car_ind)],
                           x_tail,y_tail,x_nose,y_nose)

    def move_car_show_ghost(self,pos):
        x_tail,y_tail,x_nose,y_nose,car_ind=pos
        self.canvas.coords(self.car_ids[int(car_ind)],
                           x_tail,y_tail,x_nose,y_nose)
        
        car=self.netsim.cars.objs[int(car_ind)]

        # plot ghosts arrows
        if car.ghost_infront:
            x_tailg,y_tailg,x_noseg,y_noseg,car_indg=self.pos[\
                    car.ghost_infront.ind]
            
            self.canvas.create_line([x_tail,y_tail,x_noseg,y_noseg],
                                    arrow='last',
                                    fill='#808080',
                                    width=1,
                                    tag='ghosts_pointer')
        # plot leader arrows
        #if car.car_infront:
        #    x_tailg,y_tailg,x_noseg,y_noseg,car_indg=self.pos[\
        #            car.car_infront.ind]
        #    
        #    self.canvas.create_line([x_tail,y_tail,x_noseg,y_noseg],
        #                            arrow='last',
        #                            fill='cyan',
        #                            width=1,
        #                            tag='ghosts_pointer')
            

    def plot_track(self):
        for id in self.sim.objs.keys():
            id_data=self.sim.split_id(id)
            if (id_data['category']=='track'):
                for elem in self.sim.objs[id].mods.values():
                    elem.plot()
                    
    def move(self,what_to_where,noscale=0):
        if noscale:
            for id,x_from, x_to,y_from, y_to in what_to_where:
                self.canvas.move(id, x_to-x_from , y_to-y_from)
        else:    
            s=self.canvas_scale           
            for id,x_from, x_to,y_from, y_to in what_to_where:
                self.canvas.move(id,(x_to-x_from)*s,(y_to-y_from)*s)
            
        self.adjust_area(id)


    def adjust_area(self,tag):
        """
        Adjust the area in which the entire network is built. The adjustment
        is done according to a newly placed canvas item with a tag. 
        """
        bbox=self.canvas.bbox(tag)
        
        
        resize=0
        if bbox[0]<self.area[0]:
            self.area[0]=bbox[0]-self.scrollsteps
            resize=1

        if bbox[2]>self.area[2]:
            self.area[2]=bbox[2]+self.scrollsteps
            resize=1

        if bbox[1]<self.area[1]:
            self.area[1]=bbox[1]-self.scrollsteps
            resize=1

        if bbox[3]>self.area[3]:
            self.area[3]=bbox[3]+self.scrollsteps
            resize=1

        if resize:
            #print 'bbox=',bbox,self.area
            self.set_area()
            
        
    def set_area(self):
        """
        Explicitely set area size in which the network is constructed.
        
        """
        self.canvas.coords('area',
                           self.area[0],self.area[1],
                           self.area[2],self.area[3])
            
        self.scrolledcanvas.resizescrollregion()

    def magnify_xy(self,x,y,origin_x=0,origin_y=0):
        """
        Returns transformed, magnified coordinates taking
        coordinates x,y, at 100% enlargement, and using netmons
        current magnification and origin (0,0), if not specified.
        """
        s=self.canvas_scale
        return x*s+origin_x, y*s+origin_y

    def unmagnify_xy(self,x,y,origin_x=0,origin_y=0):
        """
        Returns coordinates with 100% magnification, taking magnified,
        coordinates x,y and using netmons
        current magnification and origin (0,0), if not specified.
        """
        s=self.canvas_scale
        return (x-origin_x)/s, (y-origin_y)/s
        
    def get_unmagnified_canvas_coords(self,event):
        """
        Returns the coordinates x,y from the netmon canvas event at
        100% magnification.
        """
        return self.unmagnify_xy(self.canvas.canvasx(event.x),
                                 self.canvas.canvasy(event.y))
    
    def magnify(self,mag=100,zoom=0,origin_x=0,origin_y=0):
        """
        Magnify canvas to absolute magnification mag or zoom from present
        magnification. Both, mag and zoom are given in %.
        """
        
        if zoom>0:
            canvas_scale_new=0.01*zoom*self.canvas_scale
            if (canvas_scale_new>0.01)&(canvas_scale_new<10):
                scale=0.01*zoom
                self.canvas_scale=canvas_scale_new
            else:
                scale=1
        else:
            canvas_scale_new=0.01*mag
            if (canvas_scale_new>0.01)&(canvas_scale_new<10):
                scale=0.01*mag/self.canvas_scale
                self.canvas_scale=canvas_scale_new
            else:
                scale=1

        # rescale the coordinated of all canvas objects
        self.canvas.scale('all',origin_x,origin_y, scale, scale)
        #print 'canvas_origin_x,canvas_origin_y,canvas_scale=',\
        #      self.canvas_origin_x,self.canvas_origin_y,self.canvas_scale

        # scale width of all cavas objects with tag  width=x.x
        for n in range(2,7):
            width=0.5*n
            self.canvas.itemconfig('width=%.1f'%(width),
                                   width=width*self.canvas_scale)
        # replot maps with current scale
        for map in self.maps.values():
            self.plot_map(map)

        self.adjust_area('all')

        # inform simulation objects about rescaling
        for obj in self.sim.objs.values():
            obj.rescaled(self)

    def scale_obj(self,tags,origin_x=0,origin_y=0):
        """
        Put tagged object(s) into netmons scale.
        Currently there is NO offset scale. Scale center is always 0,0
        """
        
        if self.canvas_scale!=1:
            self.canvas.scale(tags,origin_x,origin_y,
                              self.canvas_scale,self.canvas_scale)
            
    def tag_category(self,id):
        """
        Add id to the tag category of the corresponding module. 
        """
        #if id:
        #    category=self.sim.split_id(id)['category']
        #    if category:
        #        print 'tag_category',category,id
        #        self.canvas.addtag_withtag(category,id)
        pass
    
    def find_map(self,event):
        """
        Finds and returns map under cursor
        """
        obj,tag=self.find_obj(event,map_flag=1)
        #print 'find_map',tag
        if self.maps.has_key(tag):
            return self.maps[tag]
        else:
            return None
        
    def find_obj(self,event,map_flag=0):
        """
        Return object closest to the cursor-position.
        The returned tag is the tag near the pointer with
        the most detailled information i.e. if there is a part
        of the object then the tag of the part is returned.

        If map_flag is true then only returns only background maps
        are returned.
        """
        #print 'windowcoords=',event.x,event.y
        x=event.widget.canvasx(event.x)
        y=event.widget.canvasy(event.y)
        #x=self.canvas.canvasx(event.x)
        #y=self.canvas.canvasy(event.y)
        
            
        return self.find_obj_at_coords(x,y,map_flag) 
    
    def find_obj_at_coords(self,x,y,map_flag=0):
        """
        returns simulation object and tag under
        the canvas coordinates x,y

        If map_flag is true then only returns only background maps
        are returned.
        """

        # define small box in which tags are searched
        
        canvasobj_ids=self.canvas.find_overlapping(\
            x-self.sim.node_attract/2.0,
            y-self.sim.node_attract/2.0,
            x+self.sim.node_attract/2.0,
            y+self.sim.node_attract/2.0)

        #allids=self.canvas.find_withtag('all')
        
        #print 'find_obj_at_coords  in',\
        #      [x-self.sim.node_attract/2.0,
        #       y-self.sim.node_attract/2.0,
        #       x+self.sim.node_attract/2.0,
        #       y+self.sim.node_attract/2.0]
        
        #for id in allids:
        #    tags=self.canvas.gettags(id)
        #    print '  id',id
        #    for tag in tags:
        #        print '    tag,bbox',tag,self.canvas.bbox(tag)
            
            
        invalid=1
        ind=0
        obj=None
        thetag=''
                    
        for canvasobj_id in canvasobj_ids:
            tags=self.canvas.gettags(canvasobj_id)
            #print 'find_obj_at_coords id, tags=',canvasobj_id,tags
            
            if len(tags)>0:
                for tag in tags:
                    # quick check if canvas object
                    # is a valid simulation object
                    if len(string.split(str(tag),self.sim.id_sep))>1:
                        if not obj:
                            obj=self.sim.split_id(tag)['obj']
                            # consider netmon stuff only when map_flag is on
                            if not map_flag:
                                if obj.id==self.id:
                                    obj=None
                                else:
                                    thetag=tag
                            else:
                                if obj.id==self.id:
                                    thetag=tag
                        # try to get a longer, more refined tag of same object 
                        elif (obj==self.sim.split_id(
                                tag)['obj'])&\
                                (len(tag)>len(thetag)):
                                thetag=tag
        #if thetag=='':
        #    print 'find_obj_at_coords empty thetag:',obj, thetag
        #else:
        #    print 'find_obj_at_coords obj.type,thetag:',obj.type, thetag
                
        return obj,thetag
        
    def delete(self,ids):
        for id in ids:
            self.canvas.delete(id)

    

    def del_mod(self,mod):
        self.delete([mod.id])

    def put_mod(self,mod):
        tags_found=self.canvas.find_withtag(mod.id)
        if tags_found:
            # print 'Netmon: refresh',mod.id
            self.del_mod(mod)
        self.put(mod.obj.get_plot_items(mod))


    #-------------------------------------------------------------------------
    # methods forwarded to canvas, but scales if necessary

    #def (self,**args):
    #    self.canvas.(args)

    def addtag_withtag(self,tags,tag):
        return self.canvas.addtag_withtag(tags,tag)
    
    def lower(self,tags):
        return self.canvas.lower(tags)

    def move_on_canvas(self,tag=-1,dx=0,dy=0,noscale=0):
        if noscale:
            return self.canvas.move_canvas(tag,dx,dy)
        else:
            return self.canvas.move(tag,
                                    dx*self.canvas_scale,
                                    dy*self.canvas_scale)
    

    def canvasx(self,**args):
        return self.canvas.canvasx(args)

    def canvasy(self,**args):
        return self.canvas.canvasy(args)

    def create_oval(self,x1,y1,x2,y2,**args):
        #s=self.canvas_scale
        s=1.0
        id=self.canvas.create_oval(x1*s,y1*s,x2*s,y2*s,**args)
        #self.tag_category(tag)
        self.scale_obj(id)
        return id
        
    def create_line(self,coords,**args):
        #print 'create_line,coords,tag ',coords,tag
        id=self.canvas.create_line(coords,**args)
        #self.tag_category(tag)
        self.scale_obj(id)
        return id

    def create_text(self,coords,**args):
        id= self.canvas.create_text(coords,**args)
        #self.tag_category(tag)
        self.scale_obj(id)
        return id

    def delete(self,tags):
        return self.canvas.delete(tags)
        
    def itemconfigure(self,tag,**args):
        return self.canvas.itemconfigure(tag,**args)
        
    def find_withtag(self,tags):
        return self.canvas.find_withtag(tags)
