#from classlib.track import *
from classlib.base import *
from classlib.man import *
from copy import copy
from RandomArray import *
from string import split,atoi

class Transport(Task):
    """ 
    Task to transport a item in a carrier to his/her destination.
    1. waiting untill a carrier of this management is
       boarding with this item
    2. starting a process to schedule the respective carrier.
    3. starting a process to inform all involved track managements about
       the destination of this carrier.
    """
    
    def init_task(self,args):
        """
        Read and store information about the item to be transported.
        """
        
        # store  item and dest
        self.item_id=args['item_id']
        self.dest_id=args['dest_id']
        self.orig_id=''
        self.car=None # pointer to carrier dir of carman

        # carrier will be assigned later...
        car_id=''
        self.cartriptime_start=0.0
        
        if args.has_key('car_id'): # is carrier already known ?
            car_id=args['car_id']
        # is there already a carrier assigned to this item ?
        elif self.obj.car_search_task.has_key(self.item_id):
            car_id=self.obj.car_search_task[self.item_id]['id']
            del self.obj.car_search_task[self.item_id]
            
        mode=args.get('mode','loaded')

        #print 'init_task',self.id,'init for',args['item_id'],self.dest_id,mode

        # print '  found carrier',car_id
        if car_id:
            # schedule carrier immediately, if carrier is known
            if self.obj.cars.has_key(car_id):
                
                # if carrier is empty, check whether the carrier
                # should search for items to load on its way to destination ?
                self.schedule(self.obj.cars[car_id],mode)
                #print  'scheduled',self.obj.cars[car_id]['id'],'at',\
                #      self.obj.cars[car_id]['elem_id']
            else:
                print
                print  'WARNING init_task',cars[car_id],'not in database!?'
                
        else:
            #print '  No car'
            #print '  Currenly there is no carrier assigned to this item'
            # inscribe to table with items searching for carriers
            #print 'Task',self.id,': put',self.item_id,'in task_search_car'
            self.obj.task_search_car[self.item_id]=self
            # set mode
            self.mode=mode
            # Now wait untill the schedule method is called
            # by the management when a carrier is found.

    
        
    def schedule(self,car,mode='loaded'):
        """
        Schedule the carrier car.

        If search_item is set then try to load an item on the way
        to its destination.
        """
        
        
        # print self.obj.type,'task',self.id,'.schedule:',car['id']
        # now task starts actually to do something:
        self.orig_id=car['elem_id']
        
        #print  'schedule',self.id,car['id'],\
        #      '\n  from',self.orig_id,'to',self.dest_id
        
        # check out if this carrier has not another task to serve
        # kill this task, if necessary.
        # in a future version with scheduling db
        # this must be done with more care
        if car.has_key('task'):
            # just to be sure not to kill itself,
            # neither a task that has been ordered by another management
            oldtask=car['task']
            if (oldtask!=self)&(oldtask.client_id==self.obj.id): 
                #print '  kill task:',car['task'].id
                # clean up already here to avoid double allocation
                # because it may take one more iteration to
                # cleanup old task
                if oldtask.mode=='search_item':
                    self.obj.remove_idlecar(car['id'],oldtask.orig_id)
                    #print '  remove_idlecar from',oldtask.orig_id,'\n',\
                    #          self.obj.stops[oldtask.orig_id]['idlecars']
                else: 
                    self.obj.remove_sentcar(car['id'],oldtask.dest_id)
                    #print '  remove_sentcar from',oldtask.dest_id,'\n',\
                    #          oldtask.obj.stops[self.dest_id]['sentcars']

                #if (oldtask.mode=='search_item')&(mode!='search_item'):
                #    self.obj.remove_idlecar(car['id'],self.orig_id)
                #elif (oldtask.mode=='empty')&(mode!='empty'):
                #    self.obj.remove_idlecar(car['id'],self.orig_id)
                car['task'].kill()

        if mode=='search_item':
            #self.obj.idletasks[self.id]=self
            self.obj.stops[self.orig_id]['idlecars'].append(car['id'])
        else: #elif mode=='empty':
            self.obj.stops[self.dest_id]['sentcars'].append(car['id'])

        
        
        # add info that gets passed to the ordering management 
        self.info['car_id']=car['id']

        # set processing status
        self.set_status('processing')

        # return explicitely info that task found a carrier to transport item
        self.return_info()

        
        
        # set arrived flag
        self.arrived=0

        # set mode
        self.mode=mode
        
        # initial schedule process
        self.sched_car=None
        
        # delete item from table with items searching for carriers
        #print 'Task',self.id,': remove',self.item_id,'from task_search_car'
        #print '     and assign to',car['id']
        if self.obj.task_search_car.has_key(self.item_id):
            del self.obj.task_search_car[self.item_id]

        
        
        # store link to carrier for this task
        self.car=car

        # and back-link from there, so that carrier management
        # can check whether carrier has already a task
        # later this must be refined with scheduling data
        car['task']=self

        #print '!!!!!!!!!!!!!!!!!!! store start of carrier trip time',car['id']
        #self.info['cartriptime_start']=self.obj.sim.time
        self.cartriptime_start=self.obj.sim.time
        # store start of trip if an item is loaded
        if self.car_loaded():
            self.info['triptime_start']=self.obj.sim.time
            

        # start scheduling process of resource
        self.sched_car=self.add_proc(Sched_car)
        
        

        # make an order to all involved track managements to
        # transport the respective carrier towards its destination
        # This task involves global routing over several track clusters,
        # which is currently not implemented.
        # Instead, a processes are started to inform all (the only)
        # track cluster about the destination of the carrier.
        
        # the global information is currently obtained from the
        # simulation environment (which will not exist in reality),
        # instead, a network management needs to be implemented...
        
        # go through all available simulation object types (=keys)
        trackmanproc=None
        for obj_type in self.obj.sim.objs.keys():
            if not trackmanproc:
                if string.find(obj_type,'trackman')!= -1: # is it a trackman ?
                    # start transport process for this trackman
                    #print '  add_proc Trans_car for carman',obj_type
                    trackmanproc=self.add_proc(Trans_car,obj_type)

    def car_loaded(self):
        """
        Helper functions that returns 1 if a carrier is associated with this
        task and the carrier is loaded with an item.
        """
        if self.car:
            #print 'car_loaded:',self.car['id'],len(self.item_id)
            return (len(self.item_id)>0)&(self.item_id!='empty')
        else:
            #print 'car_loaded: no carrier'
            return 0
            

            
        
    def process_info(self,info):
        """
        Process newly received info from process.
        These may be new trip estimates. 
        """
        
        if info.has_key('trippath_est'):
            #print 'ccccccccccccccccccc trippath_est=',info['trippath_est']
            #print 'ccccccccccccccccccc triptime_est=',info['triptime_est']
            self.info.update(info.copy())
            # return info to client process
            self.return_info(info)
            
    def update(self,proc):
        """
        Check whether process has been completed successfully,
        i.g if item arrived at its destination. If yes
        kill all other processes.
        """
        

        # check termination condition: schedule process must be finished
        # and at least one of the track clusters must report the arrival of car
        # or schedule process is done and carrier searches for an item.
        if self.sched_car:
            #print self.obj.id,':update task',self.id,'.',proc.id
            #if proc.status!='done':
                #print '===> ',proc.order_id,proc.status,\
                #      'done for',self.item_id,self.car['id']
                #print '     Destination:',self.dest_id
                
            if proc.status=='done':
                # carrier schedule task done.
                if proc==self.sched_car: 
                    #print '   ',self.car['id'],\
                    #      'is halted on a berth and unloading'

                    # update db with current carrier position
                    self.car['sec_id']=proc.info['sec_id']
                    self.car['elem_id']=proc.info['elem_id']

                    # if carrier found an item before regular destination
                    # was reached, set arrived flag to 1
                    # this means all remaining tasks of the original path
                    # will be removed.
                    if proc.info.has_key('found_item'):
                        if proc.info['found_item']:
                            # set 'arrived', don't wait for track to terminate
                            self.arrived = 1
                    
                            
                else:   
                    #print '  trackman schedule task done for',self.car['id']
                    self.arrived = 1

                #if (proc==self.sched_car) & self.search_item:
                #    if proc.info.has_key('elem_id'):
                #        if proc.info['elem_id']!= self.dest_id:
                #            print '!!!! detected early stop '
                #            self.arrived = 1
                    
                #
                #print '  sched_car=',self.sched_car.id,\
                #      'arrived=',self.arrived
                
                if (self.sched_car.status=='done') &  self.arrived:
                    
                
                    #if self.search_item:
                    #print self.obj.id,':ccccccccccccccccccccccccccccccccccccc'
                    #print self.__dict__
                    #print self.car['id'],\
                    #      'arrived at',self.dest_id,\
                    #      self.search_item
                    
                    self.do_accounting()
                    
                    # cancel remaining orders previously
                    # given to other track mans
                    # print '  cancel remaining procs'
                    for proc_to_kill in self.procs.values():
                        if proc_to_kill!=proc:
                            #print 'start killing',proc_to_kill.id
                            proc_to_kill.kill()
            
    def do_accounting_specific(self):
        """
        Do task specific accounting.
        To be overwritten
        """
            
        # do statistics
        self.car['trips']+=1
        
        if self.car_loaded():
            self.car['trips_loaded']+=1
            # add trip time in h
            self.car['time_loaded']+=(self.obj.sim.time-
                                      self.cartriptime_start)/3600.0
        
    def cleanup(self):
        """
        Remove all memory traces of this task.
        """
        #print 'task.cleanup',self.id,'for',self.car['id'],self.mode
        #print self.obj.id,':task.cleanup',self.id,'for',self.car['id']

        
        
        
                #self.obj.stops[self.dest_id]['sentcars'].remove(self.car['id'])
                #for ind in range(\
                #self.obj.stops[self.dest_id][\
                #'sentcars'].count(self.car['id'])):
                #    self.obj.stops[self.dest_id][\
                #    'sentcars'].remove(self.car['id'])
        
        # remove pointer from car to this task
        if self.car.has_key('task'):
            if self.car['task']==self: # is it really a pointer to this task ?
                # do empty mission stuff
                if self.dest_id:
                    if self.mode=='search_item':
                        self.obj.remove_idlecar(self.car['id'],self.orig_id)
                        #print '  remove_idlecar from',self.orig_id,'\n',\
                        #      self.obj.stops[self.orig_id]['idlecars']
                    else: #elif self.mode=='empty':
                        #print '  remove_sentcar from',self.dest_id,'\n',\
                        #      self.obj.stops[self.dest_id]['sentcars']
                        self.obj.remove_sentcar(self.car['id'],self.dest_id)

                # finally remove itselfas task associated with car
                del self.car['task']

        # delete item from table with items searching for carriers
        # if still there.
        if self.obj.task_search_car.has_key(self.item_id):
            if self.obj.task_search_car[self.item_id]==self:
                #print 'Del Task',self.id,\
                #      ': remove',self.item_id,'from task_search_car'
                del self.obj.task_search_car[self.item_id]
            
class Sched_car(Proc):
    """
    Process that orders to schedule the task's carrier to transport
    and, if existant, to unload an item at its destination.
    """

    def init_proc(self):
        """
        Make a transport order to the task's carrier.
        """
        # here the carrier should be scheduled for an expected time frame
        # (currently not implemented)

        # order carrier to move
        self.make_order(self.task.car['id'],'transport',
                        {'dest_id':self.task.dest_id,
                         'item_id':self.task.item_id,
                         'mode':self.task.mode
                         })

        self.set_status('processing')

        
        

class Trans_car(Proc):
    """
    Process that orders a track management to transport 
    a carrier to its destination.
    """

    def init_proc(self,trackman_type):
        """
        Make a transport order to this track management.
        """
        # print 'init_proc: schedule trackman with',self.task.car['id'],\
        # 'on sec',self.task.car['sec_id'],'dest',self.task.dest_id
        self.set_status('processing')
        self.make_order(trackman_type,'transport',
                        {'car_id':self.task.car['id'],
                         'sec_id':self.task.car['sec_id'], # current section
                         'dest_id':self.task.dest_id, # destination by task
                         'mode':self.task.mode
                         })
        
        self.trackman_type=trackman_type # store management to deal with

    def process_info(self,info):
        """
        Process newly received info
        from resource.
        These may be new trip estimates. 
        """
        
        if info.has_key('trippath_est'):
            self.task.put_info(info)
 
class Experim(Man):
    """
    Carrier management simulation object.
    (Expetimental version)
    """
    def init_obj_specific(self):
        self.__version__=self.sim.__version__

        self.name='Experimental carrier management'
        self.info='Generic carrier management.'
        
        # databases

        # car_id -> car data
        self.cars=self.resources
        

        # tasks with items searching for a carrier item_id -> task
        self.task_search_car={}

        # carriers with items searching for a task item_id -> car
        self.car_search_task={}
        # add operations that can be executed by an incoming message
        #self.mes_ops['add_car']=self.add_car
        #self.mes_ops['del_car']=self.del_car
        self.mes_ops['update_car']=self.update_car
        self.mes_ops['load_car']=self.load_car
        self.mes_ops['idle_car']=self.idle_car
        #self.mes_ops['put_dest']=self.put_dest
        self.mes_ops['unknown_dest']=self.unknown_dest

        # The table task_classes is mapping task names to task classes.
        self.task_classes['transport']=Transport
        
        # initialise destinations, mapping destination id to track man id
        #self.dests={}
        self.stops={}
        self.ids_maintenance=[]
        
        # poll_destinations will be a permanent task which is automatically
        # updating self.dests  by its clients i.e. the track managements
        #self.poll_dests_task_id='' # will be done when first user is added
                                   # or during set_state with saved data
                                   

        # empty vehicle stuff                           
        #self.emptycars_sentto={} # key=dest.id, value=list of carrier ids
        self.weight_queuing=0.9
        self.weight_dist=0.1

    def get_state(self):
        '''Prepare state image for save operation'''
        state_image=self.get_state_man({})
        state_image['cars']=self.cars
        #state_image['dests']=self.dests
        state_image['stops']=self.stops
        state_image['ids_maintenance']=self.ids_maintenance
        #state_image['poll_dests_task_id']=self.poll_dests_task_id
        state_image['task_search_car']=self.task_search_car
        state_image['car_search_task']=self.car_search_task

        #state_image['emptycars_sentto']=self.emptycars_sentto
        state_image['weight_queuing']=self.weight_queuing
        state_image['weight_dist']=self.weight_dist
        #state_image['']=self.
        return state_image
        
    def set_state(self,state_image):
        '''Set state while opening simulation file'''
        self.set_state_man(state_image)
        self.cars=state_image['cars']
        

        # get id from poll_dests_task
        # poll_destinations will be a permanent task which is automatically
        # updating self.dests  by its clients i.e. the track managements
        #self.poll_dests_task_id=state_image['poll_dests_task_id']
        
        self.task_search_car=state_image['task_search_car']
        self.car_search_task=state_image['car_search_task']

        if state_image.has_key('weight_queuing'):
            #self.emptycars_sentto=state_image['emptycars_sentto']
            self.weight_queuing=state_image['weight_queuing']
            self.weight_dist=state_image['weight_dist']
            #self.dests=state_image['dests']
            self.stops=state_image['stops']
            self.ids_maintenance=state_image['ids_maintenance']
        else:
            # no empty carrier management
            # update state accordingly
            for task in self.tasks.values():
                if hasattr(task,'search_item'):
                    task.orig_id=''
                    if task.item_id=='empty':
                        task.mode='search_item'
                    else:
                        task.mode='loaded'
                    
        #self.=state_image['']

    def link(self):
        self.trackman=self.try_link_with('test.trackman.experim')
        
    def get_resource_name(self):
        """
        Returns string with name of resource that is managed by this management
        To be overwritten.
        """
        return 'Carrier'

    def reset_statistics_specific(self):
        """
        Reset or initialize statistical variables specific to
        a particular type.

        To be overwritten.
        """
        # do through resources here as this function may be called
        # from the general management class
        #print 'reset_statistics_specific for',self.id
        
        for resource in self.resources.values():
            self.reset_statistics_resource(resource)

    def reset_statistics_resource(self,car):
        """
        Initialize and reset all statistical data for this carrier. 
        """
        #print 'reset_statistics_resource for ',car['id']
        car['trips']=0
        car['trips_loaded']=0
        car['trips_loaded_rel']=0.0
        # currently no stats on path...
        #car['path']=0.0
        #car['path_loaded']=0.0
        #car['path_loaded_rel']=0.0
        car['time_loaded']=0.0
        car['time_loaded_rel']=0.0

    def make_panel_specific(self,panel):
        """
        Add track element specific panel infos.
        To be overwritten.
        """
        # add a parameters page
        panel.make_page('Parameters',self.make_parampage,self.update_parampage)
        
    def format_statistics_specific(self,descriptor,dataset):
        """
        Format management specific contents for statistical data.
        """
        descriptor_specific=[
            {'name':'Total # of trips',
             'key':'trips_total',
             'type':'integer',
             'printformat':'%9d',
             'valid':{'min':0,'max':10**8},
             'mode':'read',
            'width':0.2,
             },
            {'name':'Total # of trips with load',
             'key':'trips_total_loaded',
             'type':'integer',
             'printformat':'%9d',
             'valid':{'min':0,'max':10**8},
             'mode':'read',
            'width':0.2,
             },
            {'name':'Rel. # of trips with load',
             'key':'trips_total_loaded_rel',
             'type':'real',
             'printformat':'%3.2f',
             'valid':{'min':0,'max':10**2},
             'mode':'read',
             'width':0.2,
             'units':{'%':1}
             },
            # currently no stats on path...
            #{'name','Total path'
            # 'key':'path_total',
            # 'type':'real',
            # 'printformat':'%8.2f',
            # 'valid':{'min':0,'max':10**8},
            # 'mode':'read',
            # 'width':0.2,
            # 'units':{'km':1}
            # },
            #{'name','Total path with load'
            # 'key':'path_total_loaded',
            # 'type':'real',
            # 'printformat':'%8.2f',
            # 'valid':{'min':0,'max':10**8},
            # 'mode':'read',
            # 'width':0.2,
            # 'units':{'km':1}
            # },
            #{'name','Rel. path with load'
            # 'key':'path_total_loaded_rel',
            # 'type':'real',
            # 'printformat':'%3.2f',
            # 'valid':{'min':0,'max':10**2},
            # 'mode':'read',
            # 'width':0.2,
            # 'units':{'%':1}
            # },
            {'name':'Average time with load',
             'key':'time_mean_loaded',
             'type':'real',
             'printformat':'%8.2f',
             'valid':{'min':0,'max':10**6},
             'mode':'read',
             'width':0.2,
             'units':{'h':1}
             },
            {'name':'Rel. average time with load',
             'key':'time_mean_loaded_rel',
             'type':'real',
             'printformat':'%3.2f',
             'valid':{'min':0,'max':10**2},
             'mode':'read',
             'width':0.2,
             'units':{'%':1}
             },
            ]

        descriptor+=descriptor_specific

        n=0
        trips_total=0
        trips_total_loaded=0
        #path_total=0.0
        #path_total_loaded=0.0
        time_total_loaded=0.0
        
        # calculating statistics over all carriers
        for car in self.cars.values():
            if car['trips']>0: # count only cars with at least one trip
                n+=1
                trips_total+=car['trips']
                trips_total_loaded+=car['trips_loaded']

                # currently no stats on path...
                #path_total+=car['path']
                #path_total_loaded+=car['path_loaded']
                time_total_loaded+=car['time_loaded']

                
        dataset['trips_total']=trips_total
        
        dataset['trips_total_loaded']=trips_total_loaded
        if trips_total>0:
            dataset['trips_total_loaded_rel']=100.0*trips_total_loaded/\
                                               trips_total
        else:
            dataset['trips_total_loaded_rel']=0.0

        # currently no stats on path...
        #dataset['path_total']=path_total/1000 # converted to km
        #dataset['path_total_loaded']=path_total_loaded/1000 # converted to km
        #if path_total>0:
        #    dataset['path_total_loaded_rel']=path_total_loaded/\
        #                                       path_total*100
        #else:
        #    dataset['path_total_loaded_rel']=0.0
        if n>0:
            dataset['time_mean_loaded']=time_total_loaded/n #in hours
        else:
            dataset['time_mean_loaded']=0.0
            
        if self.sim.time>0:
            dataset['time_mean_loaded_rel']=100.0*\
                                             dataset['time_mean_loaded']/\
                                             (self.sim.time/3600)
        else:
            dataset['time_mean_loaded_rel']=0.0

        
    def format_resources_specific(self,header,table):
        """
        Updates all management specific widgets on resource page and table.
        """
        header.append({'key':'trips',
                       'name':'trips',
                       'type':'integer',
                       'printformat':'%d',
                       'valid':{'min':0,'max':10**8},
                       'mode':'read',
                       'width':0.2,
                       })

        header.append({'key':'trips_loaded',
                       'name':'trips loaded',
                       'type':'integer',
                       'printformat':'%d',
                       'valid':{'min':0,'max':10**8},
                       'mode':'read',
                       'width':0.5,
                       })

        header.append({'key':'trips_loaded_rel',
                       'name':'trips loaded',
                       'type':'real',
                       'printformat':'%.2f',
                       'valid':{'min':0,'max':10**2},
                       'units':{'%':1},
                       'mode':'read',
                       'width':0.5,
                       })
        # currently no stats on path...
        #header.append({'name','path',
        #               'key':'path',
        #               'type':'real',
        #               'printformat':'%8.2f',
        #               'valid':{'min':0,'max':10**8},
        #               'mode':'read',
        #               'width':0.2,
        #               'units':{'km':1}
        #               })
        #header.append({'name','path loaded',
        #               'key':'path_loaded',
        #               'type':'real',
        #               'printformat':'%8.2f',
        #               'valid':{'min':0,'max':10**8},
        #               'mode':'read',
        #               'width':0.2,
        #               'units':{'km':1}
        #               })
        #
        #header.append({'name','path loaded',
        #               'key':'path_loaded_rel',
        #               'type':'real',
        #               'printformat':'%3.2f',
        #               'valid':{'min':0,'max':10**2},
        #               'mode':'read',
        #               'width':0.2,
        #               'units':{'%':1}
        #               })

        
        header.append({'name':'time loaded',
                       'key':'time_loaded',
                       'type':'real',
                       'printformat':'%8.2f',
                       'valid':{'min':0,'max':10**6},
                       'mode':'read',
                       'width':0.6,
                       'units':{'h':1}
                       })
        header.append({'name':'time loaded',
                       'key':'time_loaded_rel',
                       'type':'real',
                       'printformat':'%3.2f',
                       'valid':{'min':0,'max':10**2},
                       'mode':'read',
                       'width':0.6,
                       'units':{'%':1}
                       })

        
        
        # do necessary computations for extra columns
        for row in table:
            if row['trips']>0:
                row['trips_loaded_rel']=100.0*row['trips_loaded']/\
                                         row['trips']
            else:
                row['trips_loaded_rel']=0.0

            if self.sim.time>0:
                row['time_loaded_rel']=100.0*row['time_loaded']/\
                                        (self.sim.time/3600)
            else:
                row['time_loaded_rel']=0.0

    

    def format_params(self):
        """
        Returns describtor and dataset of parameters.
        """
        descriptors=[
            {'key':'weight_queuing',
             'name':'Empty vehicle man: queuing factor',
             'type':'real',
             'printformat':'%1.2f',
             'valid':{'min':0,'max':10**3},
             'mode':'write',
             'width':0.4,
            },
            {'key':'weight_dist',
             'name':'Empty vehicle man:  distance factor',
             'type':'real',
             'printformat':'%1.2f',
             'valid':{'min':0,'max':10**3},
             'mode':'write',
             'width':0.4,
            },
            ]

        dataset={'weight_queuing':self.weight_queuing,
                 'weight_dist':self.weight_dist,
                 }
        
        return descriptors,dataset
    
    def make_parampage(self,parent,widgets):
        descriptors,dataset=self.format_params()
        widgets['param_data']=gui.DatasetEditor(descriptors,dataset,
                                               parent,
                                               applyentry_command=\
                                                   self.apply_param_gui,
                                               applybutton_command=\
                                                   self.apply_param_gui)
        widgets['param_data'].pack(expand=1,fill='both')

    
    def update_parampage(self,panel,widgets):
        """
        Update of widgets on costs page. This method is called
        when cost-page needs to be updated.

        """
        descriptor,dataset=self.format_params()
        widgets['param_data'].update_contents(dataset)

    
    def apply_param_gui(self,paramdatawidget,param_data):
        self.set_params(param_data)
            
        descriptor,dataset=self.format_params()
        paramdatawidget.update_contents(dataset)

    def set_params(self,params={}):
        """
        Set parameter data of module.
        """
        self.weight_queuing=params['weight_queuing']
        self.weight_dist=dataset=params['weight_dist']
        
     
    
    def load_car(self,message):
        """
        A carrier has loaded an item. Check out whether
        it is a item of a task which is searching for a carrier.
        If so, schedule carrier for this item.
        """
        car=self.cars[message['from']]
        item_id=message['item_id']
        #print self.id,'load_car',car['id'],'with',item_id
        #print '  self.task_search_car=',self.task_search_car 
        if self.task_search_car.has_key(item_id):
            # tell respective task to schedule carrier with item
            self.task_search_car[item_id].schedule(car)
        else:
            # What happend is that the item is already at the carrier
            # before the order for this item arrived.

            print 'WARNING in',self.id,'.load_car:',item_id,'not found?'
            print '        => put ',item_id,' in car search task database'
            self.car_search_task[item_id]=car
            
            
            #if car.has_key('task'):
            #    if car['task']:
            #        print 'Current task of',car['id'],'=',car['task'].id
            #        print '  with item:',car['task'].item_id,\
            #              'and dest:',car['task'].dest_id

            if self.debug>=5:
                for task in self.tasks.values():
                    print 'Task',task.id,'with item:',task.item_id

        
                
    def idle_car(self,message):
        """
        A carrier is empty at a stop and cannot load an item there.
        This function puts create a task for this carrier.
        """
        

        # update carrier status
        car=self.set_car_status(message)
        id_orig=message['elem_id']

        

        #print 'idle_car:',car['id'],'at',id_orig

        # first check if there is demand at the present stop
        # if yes make the car search item and send it to a random
        # destination...this is not optimal but fast
        if self.stops[id_orig]['n_users_queuing']>\
               (0.3*self.stops[id_orig]['n_berths']):
            if len(self.ids_maintenance)>0: 
                id_maintenance=self.ids_maintenance[randint(len(self.ids_maintenance))]
            else:
                id_maintenance=self.newdest()# get random dest
            #print 'Send idle', car['id'],'pick up user',id_orig, id_maintenance 
            self.make_order('transport',
                            {'car_id':car['id'],
                             'item_id':'empty',
                             'dest_id':id_maintenance,
                             'mode':'search_item',
                             }
                            )
            # algorithm finished here
            return
        
        
        
        # get stop with the maximum number of users in the queue
        # without a present or already sent empty carrier
        # ???consider distance as well!!!
                
        #stops=self.sim.get_stops().values()

        value_queuing=array(len(self.stops)*[0.0])
        value_dist=array(len(self.stops)*[0.0])
        ind_dest=0
        ids_stops=self.stops.keys()
        for id_dest in ids_stops:
            if id_dest!=id_orig:
                # get users queuing for this dest
                # subtract number of empty cars that have been already
                # sent to this destination..
                #print ' emptycars_sent=',self.stops[id_dest]['sentcars']
                value_queuing[ind_dest]=self.stops[id_dest]['n_cars_demand']

             
                if value_queuing[ind_dest]>0:
                    # only destination with users are considered
                    # get costs to get from orig to dest
                    value_dist[ind_dest]=1.0/self.trackman.get_cost(
                        id_orig,id_dest)
            
                #print ' id=',ind_dest,' ',id_dest,' q=',value_queuing[ind_dest],' d=',value_dist[ind_dest],' e=',self.stops[id_dest]['sentcars']
            ind_dest+=1

                
        value_queuing_max=max(value_queuing)
        
        
        if value_queuing_max==0:
            # this means actually that there is no queuing user at
            # non of the destination

            # => make vehicle idle and make nearest maintenence the default
            #value_dist_max=0
            demand_max=-10**8
            id_maintenance=''
            for id_dest in self.ids_maintenance:
                if id_dest==id_orig:
                    id_maintenance=''
                    break
                else:
                    #value_dist=1.0/self.trackman.get_cost(id_orig,id_dest)
                    #if value_dist>value_dist_max:
                    #    value_dist_max=value_dist
                    #    id_maintenance=id_dest
                    
                    # send car to highest demand berth
                    # demand is usually negative because no users
                    if  self.stops[id_dest]['n_cars_demand']>demand_max:
                        demand_max=self.stops[id_dest]['n_cars_demand']
                        id_maintenance=id_dest

            if id_maintenance=='':
                # no maintenence found or already in a maintenance
                id_maintenance=self.newdest()
            #print 'Send idle', car['id'],'to maint.', id_maintenance 
            self.make_order('transport',
                        {'car_id':car['id'],
                         'item_id':'empty',
                         'dest_id':id_maintenance,
                         'mode':'search_item',
                         }
                        )
        else:
            
            # decide for a destination to send carrier
            value_dist_max=max(value_dist)
            value_queuing_max=max(value_queuing)
            #,value_dist_max
            #print '  value_queuing=\n',value_queuing#,value_queuing_max
            #print '  value_dist=\n',value_dist#,value_dist_max
            
            decision=self.weight_queuing*value_queuing/value_queuing_max+\
                      +self.weight_dist*value_dist/value_dist_max
            
            #print '  decision=',decision
            
            decision_max=max(decision)
            ind_decision=nonzero(decision==decision_max)[0]
            id_dest_decision=ids_stops[ind_decision]

            #print 'idle_car:',car['id'],'at',message['elem_id'],message['sec_id']
            
            #
            

            #print '  empty car',car['id'],' sent to',id_dest_decision,ind_decision
              
            # debug stuff
            #print 'Send idle', car['id'],'to',id_dest_decision
            #shortlist=[]
            #for id in ids_stops:
            #    shortlist.append(atoi(split(id,'/')[-1]))
            #print '  ',shortlist
            #print '  ',floor(decision)
            #print '  ^^^index^^^',ind_decision
            # => send vehicle to id_dest_decision
            self.make_order('transport',
                        {'car_id':car['id'],
                         'item_id':'empty',
                         'dest_id':id_dest_decision,
                         'mode':'empty',# mode for empty mission
                         }
                        )

            
        
    #def add_car(self,message):
    #    car_id=message['from']
    #    self.cars[car_id]={'id':car_id,
    #                       'status':'edit'}
    #    self.update_car(message)
    #    # set carrier idle -> auto create a task
    #    #self.idle_car(message)

    def register_specific(self,car,message={}):
        """
        Management specific resource registration.
        Will be initialization of statistical and cost data.
        """
        #print 'register_specific:',car['id']
        #print '  ',self.resources.keys()

        # this may move to general register function in man
        self.reset_statistics_resource(car)
        
        self.update_car(message)

        # poll_destinations now solved differently with
        # asking sim in update man
        # poll_destinations will be a permanent task which is automatically
        # updating self.dests  by its clients i.e. the track managements
        #if self.poll_dests_task_id=='':
        #    if self.debug>=5:
        #        print self.id,'.poll_destinations'
        #    self.poll_dests_task_id=self.make_order('poll_destinations')
            
    #def del_car(self,message):
    #    car_id=message['from']
    #    if self.cars.has_key(car_id):
    #        del self.cars[car_id]
    #        if len(self.cars)==0:
    #            self.sim.remove_obj(self)
                
    def update_car(self,message):
        """
        Updates carrier....should go to update resource
        """
        
        car_id=message['from']
        if self.cars.has_key(car_id):
            car=self.cars[car_id]
        else:
            print 'WARNING:',self.id,'.update_car: not in resource db:',car_id
            self.add_car(message)
            return
        
        if self.debug>=5:
            print 'carman:',self.type,' update_car:',message['from']
            print 'Status:',car['status'],'-->',message['status']

        # store new status and carrier info in local carrier db
        self.set_car_status(message)
        
        
    def newdest(self):
        '''Find a new destination, usually for a carrier without load
        (currently random).
        '''
    
        newdest_id=''
        n=len(self.stops) # number of of availlable destinations
        if n>0:
            # get new random destination
            newdest_id=self.stops.keys()[randint(n)]
            # print 'newdest:',newdest_id,'of',car['id']
            
            
        else: # there are no availlable destination in the database
            #print 'Carman.newdest: no availlable destination in the database'

            # => poll all track managements for available destinations
            #self.poll_dest()
            # create a task that polls all available destinations

            # check if poll_dests_task is working
            if self.poll_dests_task_id==-1:
                # relaunch poll_dests_task
                self.poll_dests_task_id=self.make_order('poll_destinations')

        return newdest_id
        
        
        #print 'Number of available dest =',n
            
    def unknown_dest(self,message):
        
        ''' Carrier encountered an unknown destination.  Probably the
        current destination table is outdated. Poll the track management
        of the old carrier destination for a new destination table.
        Carrier with invalid detination will be branched to a stop,
        where it will receive a new destination.
        '''
        print 'WARNING:',self.users[message['from']],\
              'encoutered unknown destination'
        
    #    car=self.cars[message['from']]
    #    if car.has_key('dest_id'):
    #        if self.dest.has_key(car['dest_id']): # key is trackman id
    #            # create a task that polls all available destinations
    #            self.make_order('poll_destinations')
    #            #self.poll_dest(self.dest[car['dest_id']])
                
        
    def set_car_status(self,message):
        '''
        Reads carrier info from message and and inserts it into carrier
        database.
        '''
        car=self.cars[message['from']]
        status=message['status']

        if car['status']!=status:
            # attention status changed:
            pass
        
        if message.has_key('elem_id'):
            elem_id=message['elem_id']
            if car.has_key('elem_id'):
                if car['elem_id']!=elem_id:
                    if car.has_key('task'):
                        task=car['task']
                        if task:
                            if task.mode=='search_item':
                                #print 'set_car_status',self.id
                                #print '  moved idlecar:',car['id'],\
                                #      'from',car['elem_id'],'to',elem_id
                                self.remove_idlecar(car['id'],task.orig_id)
                            
            car['elem_id']=message['elem_id']
            car['sec_id']=message['sec_id']
        #print self.id,':Set status:\n',car
        #if message.has_key('dest_id'):
        #    if message['dest_id']!='':
        #        car['dest_id']=message['dest_id']
        return car

    def remove_idlecar(self,car_id,elem_id):
        """
        Tries to remove car with car_id from
        idlecar list of stop with elem_id
        """
        if self.stops.has_key(elem_id):
            idlecars=self.stops[elem_id]['idlecars']
            for ind in range(idlecars.count(car_id)):
                #print 'remove_idlecar',car_id,'from idlecars of',elem_id
                idlecars.remove(car_id)

    def remove_sentcar(self,car_id,elem_id):
        """
        Tries to remove car with car_id from
        idlecar list of stop with elem_id
        """
        if self.stops.has_key(elem_id):
            sentcars=self.stops[elem_id]['sentcars']
            for ind in range(sentcars.count(car_id)):
                #print 'remove_sentcar',car_id,'from sentcars of',elem_id
                sentcars.remove(car_id)
 
            
    def poll_dest(self, trackman_id=None):
        '''Polls all track managements for currently availlable destinations'''
        # this is a network wide operation and assumes the knowledge
        # of the ids of all track managements
         
        # the global information is currently obtained from the
        # simulation environment (which will not exist in reality),
        # instead, a network management needs to be implemented...

        if trackman_id: # a specific trackman_id has been given, so poll it
            self.send_mes({'to':trackman_id,
                           'op':'get_dest'
                           })
            
        else: # poll all availavle trackmans
            # go through all available simulation object types (=keys)
            for obj_type in self.sim.objs.keys():
                if string.find(obj_type,'trackman')!= -1: # is it a trackman ?
                    # poll destinations
                    self.send_mes({'to':obj_type,
                                   'op':'get_dest'
                                   })
                    # destination will be send back,
                    # calling the put_dest method
                     
    def output(self):
        
        # read and process all messages in the message buffer
        self.read_mes()

        # here we start empty missions if there are  stops
        # with demand
        demand=[]
        offer={}
        self.ids_maintenance=[]
        for stop in self.sim.get_stops().values():
            if stop.status=='active':
                n_users_queuing=stop.n_users_queuing

                # create/update stop database
                if self.stops.has_key(stop.id):
                    # rough estimation:
                    # the vehicle offer at the berth is estimated
                    # to half of the
                    # number of vehicles currently in the berth section
                    n_cars_demand=n_users_queuing-0.5*len(stop.cars)\
                                   -len(self.stops[stop.id]['sentcars'])

                    # put stop in demand list if there is demand for carriers
                    if n_cars_demand>0:
                        demand.append((n_cars_demand,stop.id))
                    self.stops[stop.id]['n_berths']=stop.berthssec.berth_no
                    self.stops[stop.id]['n_users_queuing']=n_users_queuing
                    self.stops[stop.id]['n_cars_demand']=n_cars_demand
                

                else:
                    # rough estimation:
                    # the vehicle offer at the berth is estimated
                    # to half of the
                    # number of vehicles currently in the berth section
                    n_cars_demand=n_users_queuing-0.5*len(stop.cars)
                    self.stops[stop.id]={'n_users_queuing':n_users_queuing,
                                         'n_berths':stop.berthssec.berth_no,
                                         'sentcars':[],
                                         'idlecars':[],
                                         'n_cars_demand':n_cars_demand}

                if stop.zone=='maintenance':
                    self.ids_maintenance.append(stop.id)
                    
                # determine offer
                n_cars_offer=len(self.stops[stop.id]['idlecars'])
                if n_cars_offer>0:
                    offer[stop.id]=n_cars_offer
        

        # create an empty mission for an ideal task
        # that is closest to the stop with demand
        #demand.sort()<

        ind_demand=len(demand)
        

        #if ind_demand >0:
            #print 'output',self.id,'found ',ind_demand,' demand stops',\
            #      len(offer),'offer stops'
            #for id_stop in offer.keys():
            #    print '    offer stop',id_stop,offer[id_stop]
            #for task in self.tasks.values():
            #for car in self.cars.values():
            #    
            #    if car.has_key('task'):
            #        task=car['task']
            #        print '      ',car['id'],task.id,task.mode,task.dest_id
            #    else:
            #        print '      ',car['id'],'*** no task'
        #else:
        #    #print 'output',self.id,'no demand'

        
        # go through all demand berth, serve largest demand first.
        id_orig_found='*'
        while (ind_demand>0)&(id_orig_found!=''):
            ind_demand-=1
            n_cars_demand,id_dest=demand[ind_demand]
            

            # select closest offer stop
            cost_min=10**8
            id_car_found=''
            id_orig_found=''
            for id_orig in offer.keys():
                if id_orig!=id_dest:
                    cost=self.trackman.get_cost(id_orig,id_dest)
                    if (cost<cost_min)&\
                           (len(self.stops[id_orig]['idlecars'])>0):
                        cost_min=cost
                        id_orig_found=id_orig

            if id_orig_found:
                id_car_found=self.stops[id_orig_found]['idlecars'].pop(0)
                
                #print '\n Empty mission:  send idle car',id_car_found,\
                #      '\n   from',id_orig_found,'to', id_dest
                task=self.cars[id_car_found]['task']
                #print '   kill task',task.id,task.mode

                if task.mode=='search_item':
                    # explicit kill of idle task
                    self.cars[id_car_found]['task'].kill()

            
                    task_id=self.make_order('transport',
                                            {'car_id':id_car_found,
                                             'item_id':'empty',
                                             'dest_id':id_dest,
                                             'mode':'empty',# empty mission
                                             }
                                            )
                
                    if offer[id_orig_found]>0:
                        offer[id_orig_found]-=1
                    else:
                        # no more offer
                        del offer[id_orig_found]
                else:
                    print 'WARNING: in output: task',task.id,'is in mode',\
                          task.mode,'and not search_item'
            
    def show_info_specific(self,debuglevel=0):
        print '-------------'
        print 'Carriers:'
        for car in self.cars.values():
            print '  ',car['id'],'las notice at',car['elem_id']
            if car.has_key('task'):
                if car['task']!=None:
                    task=car['task']
                    print '    task',task.id,task.mode,task.dest_id

        print 'Stops:'
        for id_stop in  self.stops.keys():
            print '  ',id_stop,'demand=',self.stops[id_stop]['n_cars_demand'],\
                  'sent=',len(self.stops[id_stop]['sentcars']),\
                  'idle=',len(self.stops[id_stop]['idlecars'])
            for id_car in self.stops[id_stop]['sentcars']:
                print '      sentcar',id_car
            for id_car in self.stops[id_stop]['idlecars']:
                print '      idlecar',id_car
                

            

 
