from classlib.track import *



class Stop10_right_v16_ground(Stop):

    def make_track(self):
        """
        This function sets essential parameters for the specific track
        element. It needs to be called once during initialization.
        """

        # main line velocity
        velocity_main=60.0/3.6 # m/s
        clearingzone=40.0
        
        # ATTENTION!!
        # length of acceleration and brake section must be adapted to speeds
        ac=2.5
        length_brake=1/(2.0*ac)*velocity_main**2
        if length_brake<clearingzone:
          length_brake=clearingzone+1.0
        
        # velocity in berths section
        velocity_berth=10.0/3.6 # m/s
        
        

        # number of berths in berths section
        n_berth=10

        # length of one berth (must be greater than lenght of car)
        length_berth=3.8
        
        # total lengtth of berthsec
        length_berthsection=n_berth*length_berth
        
       
        
        # berth that must be kept clear in case there are no users waiting
        berths_keep_clear=3


        
      
        #######################################################################
        # topology
        addedlength=length_berthsection

        # length of parallel branch
        length_par=2.0*length_brake+addedlength
        
        addedlength=length_berthsection-15.0
        self.element=[\
            # s0 entry
            {'label':'entry',
             'shape': ( 'line', {'offset':(0.0,0.0,0.0),
                                 'length':5.0} ),
             'topo' : (1,2),
             'velocity_line':0.75*velocity_main,
             },
            # s1 main at entry
            {'label':'main_entry',
             'shape': ('line', {'offset':(5.0,0.0,0.0),
                                'length':length_par/2}),
             'topo' : (2,3),
             'velocity_line':velocity_main,
             'clearingzone':clearingzone,
             },
            # s4 main at exit
            {'label':'main_exit',
             'shape': ('line', {'offset':(5+length_par/2,0.0,0.0),
                                'length':length_par/2}),
             'topo' : (3,4),
             'velocity_line':velocity_main,
             },
            # s4 exit
            {'label':'exit',
             'shape': ('line', {'offset':(5.0+length_par,0.0,0.0),
                                'length':5.0}),
             'topo' : (4,5),
             'velocity_line':velocity_main,
             },
            # s5 brake-line
            {'label':'brake',
             'velocity_line': velocity_berth,
             'shape': ('scurve', {'offset':(5.0,5.0,0.0),
                                  'length':length_brake,
                                  'displace':-5.0,
                                  'turn'  :'right'
                                  }),
             'clearingzone':clearingzone,
             'topo' : (2,6),
             },
            # s6 berths
            {'label':'berths',
             'type':'berth',
             'berth_no':    n_berth,
             'berth_length':length_berth,
             'berths_keep_clear': berths_keep_clear,
             'shape':         ('line', {'offset':(5.0+length_brake,5.0,0),
                                        'length':addedlength}),
             'velocity_line': velocity_berth,
             'side_plat':'left',
             'topo' :         (6,7),
             },
            #s7 acceleration-line
            {'label':'accel',
             'shape': ('scurve', {'offset':(5.0+length_brake+addedlength,
                                            0.0,0.0),
                                  'length':length_brake,
                                  'displace':-5.0,
                                  'turn'  :'left'
                                  }),
             'topo' : (7,4),
             'velocity_line':velocity_main,
             },
            ]

        self.nodes_connect={1:'in', 5:'out'}


        self.info= self.info= 'Stop with 10 berths which are off-line, '+\
                   'on the right side\n'+\
                   'and ground platform on left side.\n'+\
                   'Default main line velocity is 16.6m/s\n '+\
                   'DO NOT EXCEED THIS VELOCITY !'+\
                   'Both nodes can be dragged in order to change the length.'

        
        # define class for this track element module
        self.Mod=Stopmod

 
